"""
This module contains all the functions related to wrapping NGSolve meshes to FEniCSx
We adopt the same docstring conventiona as the FEniCSx project, since this part of
the package will only be used in combination with FEniCSx.
"""

import typing
import basix.ufl
import dolfinx

try:
    from dolfinx.io import gmshio
except ImportError:
    from dolfinx.io import gmsh as gmshio

import numpy as np
import numpy.typing as npt
import ufl
from packaging.version import Version
from mpi4py import MPI as _MPI
from ngsPETSc.utils.utils import find_permutation
from ngsPETSc import MeshMapping

# Map from Netgen cell type (integer tuple) to GMSH cell type
_ngs_to_cells = {(2, 3): 2, (2, 4): 3, (3, 4): 4}


def _dim_to_element_wrapper(ngmesh: typing.Any) -> dict[int, typing.Any]:
    """Convenience wrapper to extract elements from a NetGen mesh based on topological dimension"""
    return {
        1: ngmesh.Elements1D,
        2: ngmesh.Elements2D,
        3: ngmesh.Elements3D,
    }


def _numpy_trim(array: npt.ArrayLike, dtype: npt.DTypeLike) -> np.ndarray:
    """Trim zeros from a two dimensional numpy array along axis 1
    and return it as a specified dtype.

    Args:
        array: The input array to trim.
        dtype: The desired output dtype.
    Returns:
        The trimmed array.
    """
    if Version(np.__version__) >= Version("2.2"):
        trimmed = np.trim_zeros(
            array,
            "b",
            axis=1,
        ).astype(dtype)
    else:
        trimmed = np.array(
            [list(np.trim_zeros(a, "b")) for a in list(array)],
            dtype=dtype,
        )
    return trimmed


class GeometricModel:
    """
    This class is used to wrap a Netgen geometric model to a DOLFINx mesh.
    Args:
            geo: The Netgen model
            comm: The MPI communicator to use for mesh creation
    """

    def __init__(self, geo, comm: _MPI.Comm, comm_rank: int = 0):
        self.geo = geo
        self.comm = comm
        self.comm_rank = comm_rank

    def model_to_mesh(
        self,
        hmax: float,
        gdim: int = 2,
        partitioner: typing.Callable[
            [_MPI.Comm, int, int, dolfinx.cpp.graph.AdjacencyList_int32],
            dolfinx.cpp.graph.AdjacencyList_int32,
        ] = dolfinx.mesh.create_cell_partitioner(dolfinx.mesh.GhostMode.shared_facet),
        transform: typing.Any = None,
        routine: typing.Any = None,
        meshing_options: dict[str, typing.Any] | None = None,
    ) -> tuple[
        dolfinx.mesh.Mesh,
        tuple[dolfinx.mesh.MeshTags | None, dolfinx.mesh.MeshTags | None],
        dict[tuple[int, str], tuple[int, ...]],
    ]:
        """Given a NetGen model, take all physical entities of the highest
        topological dimension and create the corresponding linear DOLFINx mesh.

        Args:
            hmax: The maximum diameter of the elements in the triangulation
            model: The NetGen model
            gdim: Geometrical dimension of the mesh
            partitioner: Function that computes the parallel
                distribution of cells across MPI ranks
            transform: PETSc DMPLEX Transformation to be applied to the mesh
            routine: Function to be applied to the mesh after generation
                takes as plan the mesh and the NetGen model and returns the
                same objects after the routine has been applied.

        Returns:
            A DOLFINx mesh for the given NetGen model. It also extracts cell tags,
            facet tags and a mapping from the NetGen label to the corresponding integer marker(s).
        """
        meshing_options = {} if meshing_options is None else meshing_options

        # To be parallel safe, we generate on all processes
        # NOTE: This might change in the future.

        # First we generate a mesh
        ngmesh = self.geo.GenerateMesh(maxh=hmax, **meshing_options)
        self.ngmesh = ngmesh
        # Apply any ngs routine post meshing
        if routine is not None:
            ngmesh, self.geo = routine(ngmesh, self.geo)
        if transform is not None:
            meshMap = MeshMapping(ngmesh)
            transform.setDM(meshMap.plex)
            transform.setUp()
            newplex = transform.apply(meshMap.plex)
            meshMap = MeshMapping(newplex)
            ngmesh = meshMap.ngmesh

        assert ngmesh.dim in (2, 3), "Only 2D and 3D meshes are supported."
        regions = self.extract_regions()
        ct, ft = self.extract_linear_mesh(gdim=gdim, partitioner=partitioner)
        return self._mesh, (ct, ft), regions

    def extract_regions(self):
        """Extract regions from the Netgen mesh."""
        ngmesh = self.ngmesh
        # Create lookup from cells/facets materials to integer tags
        regions: dict[tuple[int, str], list[int]] = {}
        # Append facet material
        for dim in (ngmesh.dim, ngmesh.dim - 1):
            for name in ngmesh.GetRegionNames(dim=dim):
                regions[(dim, name)] = []
            for i, name in enumerate(ngmesh.GetRegionNames(dim=dim)):
                regions[(dim, name)].append(i + 1)

        for key, value in regions.items():
            regions[key] = tuple(value)
        return regions

    def extract_linear_mesh(
        self,
        gdim: int,
        partitioner=dolfinx.mesh.create_cell_partitioner(
            dolfinx.mesh.GhostMode.shared_facet
        ),
    ) -> tuple[dolfinx.mesh.MeshTags, dolfinx.mesh.MeshTags]:
        """
        Extract a DOLFINx mesh (and correpsonding cell and facet tags) from the Netgen mesh.

        Args:
            gdim: Geometric dimension of the mesh
            partitioner: Function that computes the parallel distribution of cells across MPI ranks

        Note:
            This function updates the `self._mesh` to be in sync with the NetGen model.

        Returns:
            The cell and facet tags of the DOLFINx mesh.
        """
        # Extract the elements from the NetGen mesh

        ngmesh = self.ngmesh
        elements_as_numpy = _dim_to_element_wrapper(ngmesh)[ngmesh.dim]().NumPy()
        T = elements_as_numpy["nodes"]

        # Sort elements by number of vertices, i.e. group them by cell type
        number_of_vertices = elements_as_numpy["np"]
        sorted_index = np.argsort(number_of_vertices)
        self._sorted_mapping = (
            sorted_index  # Store the mapping from ngsolve to DOLFINx ordering
        )
        offset = (
            number_of_vertices[sorted_index[1:]] - number_of_vertices[sorted_index[:-1]]
        )

        type_offset = np.concatenate(
            [
                np.zeros(1, dtype=np.int64),
                np.flatnonzero(offset) + 1,
                np.array([elements_as_numpy.shape[0]], dtype=np.int64),
            ],
        ).astype(np.int64)

        # Defcide if we have a mixed mesh
        num_points_per_cell = number_of_vertices[sorted_index][type_offset[:-1]]
        if len(type_offset) == 2:
            mixed_mesh = False
        else:
            mixed_mesh = True
            assert len(type_offset) == 3, (
                "Only two different cell types are expected in a 2D grid."
            )

        # We extract topology/connectivity (T) and geometry/nodes (V) from the NetGen mesh
        if self.comm.rank == self.comm_rank:
            V = ngmesh.Coordinates()
            if mixed_mesh:
                T_sorted = T[sorted_index]
                del T
                T = []
                for i in range(len(num_points_per_cell)):
                    T_by_type = T_sorted[type_offset[i] : type_offset[i + 1]]
                    T.append(_numpy_trim(T_by_type, dtype=np.int64) - 1)
            else:
                T_by_type = T[sorted_index]
                del T
                T = _numpy_trim(T_by_type, dtype=np.int64) - 1
        else:
            V = np.zeros((0, gdim), dtype=np.float64)
            if mixed_mesh:
                T = [np.zeros((0, nv), dtype=np.int64) for nv in num_points_per_cell]
            else:
                T = np.zeros((0, number_of_vertices[0]), dtype=np.int64)

        # Permute the vertices of the NetGen mesh to match the DOLFINx ordering
        # and create the DOLFINx mesh
        if mixed_mesh:
            ufl_domain = [
                gmshio.ufl_mesh(
                    _ngs_to_cells[(gdim, Ti.shape[1])], gdim, dolfinx.default_real_type
                )
                for Ti in T
            ]
            cell_strs = []
            for udi, Ti in zip(ufl_domain, T):
                cell_str = udi.ufl_coordinate_element().cell_type.name
                cell_strs.append(cell_str)
                Ti[:, :] = Ti[
                    :,
                    dolfinx.cpp.io.perm_vtk(
                        dolfinx.mesh.to_type(cell_str), Ti.shape[1]
                    ),
                ].copy()
            c_els = [
                dolfinx.fem.coordinate_element(
                    dolfinx.mesh.to_type(cell), 1
                )._cpp_object
                for cell in cell_strs
            ]
            T = [Ti.flatten().copy() for Ti in T]
            V = V[:, :gdim].copy()

            cpp_mesh = dolfinx.cpp.mesh.create_mesh(
                self.comm,
                T,
                c_els,
                V,
                partitioner,
                2  # Maximum number of cells connected to any facet
            )
            # Wrap as Python object
            mesh = dolfinx.mesh.Mesh(cpp_mesh, domain=None)

        else:
            ufl_domain = gmshio.ufl_mesh(
                _ngs_to_cells[(gdim, T.shape[1])], gdim, dolfinx.default_real_type
            )

            cell_str = ufl_domain.ufl_coordinate_element().cell_type.name
            T = T[
                :, dolfinx.cpp.io.perm_vtk(dolfinx.mesh.to_type(cell_str), T.shape[1])
            ]
            V = V[:, :gdim].copy()
            mesh = dolfinx.mesh.create_mesh(
                self.comm,
                cells=T,
                x=V,
                e=ufl_domain,
            )
        # Extract cell and facet tags (for non-mixed meshes)
        if mixed_mesh:
            ct = None
            ft = None
        else:
            ct = extract_element_tags(self.comm_rank, ngmesh, mesh, dim=ngmesh.dim)
            ct.name = "Cell tags"
            ft = extract_element_tags(self.comm_rank, ngmesh, mesh, dim=ngmesh.dim - 1)
            ft.name = "Facet tags"

        # Attach DOLFINx mesh to the GeometricModel
        self._mesh = mesh
        return ct, ft

    def curveField(self, order: int, permutation_tol: float = 1e-8):
        """
        This method returns a curved mesh as a Firedrake function.

        :arg order: the order of the curved mesh.
        :arg permutation_tol: tolerance used to construct the permutation of the reference element.
        """
        # Extract cell map(s) from the DOLFINx mesh, to be able to generate
        # the relevant basix elements for the higher order mesh.
        try:
            num_index_maps = len(
                self._mesh.topology.index_maps(self._mesh.topology.dim)
            )
            is_mixed_mesh = num_index_maps > 1
            cells = self._mesh.topology._cpp_object.cell_types
            cell_maps = self._mesh.topology.index_maps(2)
        except AttributeError:
            is_mixed_mesh = False
            cells = [self._mesh.topology.cell_type]
            cell_maps = [self._mesh.topology.index_map(2)]

        if is_mixed_mesh and order > 2:
            raise NotImplementedError(
                "Curved mixed meshes of order > 2 are not supported."
            )

        geom_dim = self.ngmesh.dim
        assert geom_dim == self._mesh.geometry.dim, (
            "Geometrical dimension of DOLFINx mesh does not match Netgen mesh."
        )

        # Create higher order function spaces to create the coordinate element dofmap from
        elements = [
            basix.ufl.element(
                "Lagrange", dolfinx.mesh.to_string(cell), order, shape=(geom_dim,)
            )
            for cell in cells
        ]
        if is_mixed_mesh:
            orders = [
                self._mesh.geometry._cpp_object.cmaps(i).degree
                for i in range(num_index_maps)
            ]
            assert len(np.unique(orders)) == 1
            if orders[0] == order:
                return self._mesh
            elements_cpp = [
                dolfinx.cpp.fem.FiniteElement_float64(e.basix_element._e, None, True)
                for e in elements
            ]
            # NOTE: Both dofmaps have the same IndexMap, but different cell_dofs
            dofmaps = dolfinx.cpp.fem.create_dofmaps(
                self._mesh.comm, self._mesh.topology._cpp_object, elements_cpp
            )

            # Create C++ function space
            V_cpp = dolfinx.cpp.fem.FunctionSpace_float64(
                self._mesh._cpp_object, elements_cpp, dofmaps
            )
            domains = [
                ufl.Mesh(
                    basix.ufl.element(
                        "Lagrange", dolfinx.mesh.to_string(cell), 1, shape=(geom_dim,)
                    )
                )
                for cell in cells
            ]

            function_spaces = [
                dolfinx.fem.FunctionSpace(
                    dolfinx.mesh.Mesh(self._mesh._cpp_object, domain), element, V_cpp
                )
                for element, domain in zip(elements, domains)
            ]
        else:
            if self._mesh.geometry.cmap.degree == order:
                return self._mesh
            function_spaces = [dolfinx.fem.functionspace(self._mesh, elements[0])]

        # Prepare array that will host the coordinates of the curved mesh
        if is_mixed_mesh:
            # All function spaces (per cell type) shares the same index map.
            # We use the first one
            space_dm = function_spaces[0]._cpp_object.dofmaps(0)
            space_im = space_dm.index_map
            x = np.zeros(
                (space_im.size_local + space_im.num_ghosts, geom_dim), dtype=np.float64
            )
        else:
            x = function_spaces[0].tabulate_dof_coordinates()[:, :geom_dim]

        # Extract global number of cells in the NetGen mesh
        dim_to_element_getter = _dim_to_element_wrapper(self.ngmesh)
        ng_element = dim_to_element_getter[self.ngmesh.dim]
        ng_dimension = len(ng_element())  # Number of cells in NGS grid (on any rank)

        # For each cell type we compute the curved coordinates by curving the netgen mesh
        offset = 0
        for i, (imap, X_space, element) in enumerate(
            zip(cell_maps, function_spaces, elements)
        ):
            rsp = (
                element.basix_element.x
            )  # NOTE: FIX empty points in basix nanobind wrapper
            _reference_space_points = []
            for lin in rsp:
                if len(lin) != 0:
                    _reference_space_points.append(np.vstack(lin))
            reference_space_points = np.vstack(_reference_space_points)
            # Curve the mesh on rank 0 only
            num_cells = imap.size_global
            if self.comm.rank == self.comm_rank:
                # Construct numpy arrays for physical domain data
                physical_space_points = np.zeros(
                    (ng_dimension, reference_space_points.shape[0], geom_dim)
                )
                curved_space_points = np.zeros(
                    (ng_dimension, reference_space_points.shape[0], geom_dim)
                )
                self.ngmesh.Curve(1)  # Reset mesh to be linear
                self.ngmesh.CalcElementMapping(
                    reference_space_points, physical_space_points
                )
                # Cells in NGSolve that correspond to given mixed type
                local_cells = self._sorted_mapping[offset : offset + num_cells]
                physical_space_points = physical_space_points[local_cells]
                self.ngmesh.Curve(order)
                self.ngmesh.CalcElementMapping(
                    reference_space_points, curved_space_points
                )
                curved_space_points = curved_space_points[local_cells]
                curved = ng_element().NumPy()["curved"][local_cells]

                # Broadcast a boolean array identifying curved cells
                curved = self.comm.bcast(curved, root=self.comm_rank)
                physical_space_points = physical_space_points[curved]
                curved_space_points = curved_space_points[curved]
            else:
                curved = self.comm.bcast(None, root=self.comm_rank)
                # Construct numpy arrays as buffers to receive physical domain data
                ncurved = np.sum(curved)
                physical_space_points = np.zeros(
                    (ncurved, reference_space_points.shape[0], geom_dim)
                )
                curved_space_points = np.zeros(
                    (ncurved, reference_space_points.shape[0], geom_dim)
                )

            # Broadcast curved cell point data
            self.comm.Bcast(physical_space_points, root=self.comm_rank)
            self.comm.Bcast(curved_space_points, root=self.comm_rank)

            # Get coordinates of higher order space on linarized geometry
            if is_mixed_mesh:
                # Use reference space points here to push forward in FEniCSx
                cmap = self._mesh.geometry._cpp_object.cmaps(i)
                dofmap = self._mesh.geometry._cpp_object.dofmaps(i)
                coords = self._mesh.geometry.x[dofmap][:, :, :geom_dim].copy()
                space_dm = X_space._cpp_object.dofmaps(i)
                cell_node_map = space_dm.map()
                num_cells_local = imap.size_local + imap.num_ghosts
                assert num_cells_local == coords.shape[0]
                assert element.basix_element.interpolation_is_identity
                assert not X_space._cpp_object.elements(i).needs_dof_transformations
                for c, cell_coords in enumerate(coords):
                    dd = cmap.push_forward(reference_space_points, cell_coords)
                    x[cell_node_map[c], :] = dd
            else:
                num_cells_local = imap.size_local + imap.num_ghosts
                cell_node_map = X_space.dofmap.list[offset : offset + num_cells_local]

            # Determine what cells that is on this rank that should be curved
            new_coordinates = x[:, :][cell_node_map]
            if is_mixed_mesh:
                if not hasattr(
                    self._mesh.topology._cpp_object, "original_cell_indices"
                ):
                    raise RuntimeError(
                        "MixedTopology not supported with this DOLFINx",
                        f" version {dolfinx.__version__}, please update DOLFINx.",
                    )

                o_cell_index = self._mesh.topology._cpp_object.original_cell_indices[i]
                igi_to_dolfinx = np.full(len(ng_element()), -1, dtype=np.int64)
                igi_to_dolfinx[o_cell_index] = np.arange(
                    len(o_cell_index), dtype=np.int64
                )
                cells_with_curving = igi_to_dolfinx[np.flatnonzero(curved)]
                local_cells_with_curving = np.flatnonzero(cells_with_curving >= 0)
            else:
                o_cell_index = self._mesh.topology.original_cell_index
                igi_to_dolfinx = np.full(len(self._sorted_mapping), -1, dtype=np.int64)
                igi_to_dolfinx[o_cell_index] = np.arange(
                    len(o_cell_index), dtype=np.int64
                )
                cells_with_curving = igi_to_dolfinx[np.flatnonzero(curved)]
                local_cells_with_curving = np.flatnonzero(cells_with_curving >= 0)

            # Compute permutation between NetGen higher order nodes and DOLFINx
            if len(local_cells_with_curving) > 0:
                psp_local = physical_space_points[local_cells_with_curving]
                permutation = find_permutation(
                    psp_local,
                    new_coordinates[cells_with_curving][local_cells_with_curving]
                    .reshape(psp_local.shape)
                    .astype(self._mesh.geometry.x.dtype, copy=False),
                    tol=permutation_tol,
                )
            else:
                permutation = np.zeros(
                    (0, physical_space_points.shape[1]), dtype=np.int64
                )

            # Apply the permutation to each cell in turn
            if len(local_cells_with_curving) > 0:
                for ii, jj in enumerate(local_cells_with_curving):
                    curved_space_points[jj] = curved_space_points[jj][permutation[ii]]
                # Assign the curved coordinates to the dat
                x[
                    cell_node_map[cells_with_curving][
                        local_cells_with_curving
                    ].flatten(),
                    :geom_dim,
                ] = curved_space_points[local_cells_with_curving].reshape(-1, geom_dim)
            offset += num_cells

        # Create a DOLFINx mesh which shares topology with the linearized mesh
        # but has a new curved geometry.
        if is_mixed_mesh:
            topology = self._mesh.topology  # Use topology from original mesh
            c_els = [
                dolfinx.fem.coordinate_element(
                    c, order, basix.LagrangeVariant.equispaced
                )._cpp_object
                for c in cells
            ]
            X_space = function_spaces[0]
            geom_imap = X_space._cpp_object.dofmaps(0).index_map
            local_node_indices = np.arange(
                geom_imap.size_local + geom_imap.num_ghosts, dtype=np.int32
            )
            igi = geom_imap.local_to_global(local_node_indices)
            node_order = np.argsort(igi)

            _xdofs = []
            for i in range(len(cells)):
                space_dm = X_space._cpp_object.dofmaps(i)
                _xdofs.append(space_dm.map().flatten())
            xdofs = geom_imap.local_to_global(np.hstack(_xdofs).astype(np.int32))

            # Sort geometry in the same order as the sorted global dof indices
            coords = x[node_order, :geom_dim].flatten().copy()
            # Create C++ geometry
            geometry = dolfinx.cpp.mesh.create_geometry(
                topology._cpp_object,
                c_els,
                igi[node_order].copy(),
                xdofs,
                coords,
                geom_dim,
            )
            # Create DOLFINx C++ mesh
            if x.dtype == np.float64:
                cpp_mesh = dolfinx.cpp.mesh.Mesh_float64(
                    self._mesh.comm, topology._cpp_object, geometry
                )
            else:
                raise RuntimeError(f"Unsupported dtype for mesh {x.dtype}")
            ufl_domain = None
        else:
            X_space = function_spaces[0]
            # Use topology from original mesh
            topology = self._mesh.topology
            # Use geometry from function_space
            c_el = dolfinx.fem.coordinate_element(elements[0].basix_element)  #  pylint: disable=E1120
            geom_imap = X_space.dofmap.index_map
            local_node_indices = np.arange(
                geom_imap.size_local + geom_imap.num_ghosts, dtype=np.int32
            )
            igi = geom_imap.local_to_global(local_node_indices)
            geometry = dolfinx.mesh.create_geometry(
                geom_imap, cell_node_map, c_el._cpp_object, x[:, :geom_dim].copy(), igi
            )

            # Create DOLFINx C++ mesh
            if x.dtype == np.float64:
                cpp_mesh = dolfinx.cpp.mesh.Mesh_float64(
                    self._mesh.comm, topology._cpp_object, geometry._cpp_object
                )
            else:
                raise RuntimeError(f"Unsupported dtype for mesh {x.dtype}")
            ufl_domain = ufl.Mesh(elements[0])  # Use first element as domain

        # Wrap as Python object
        return dolfinx.mesh.Mesh(cpp_mesh, domain=ufl_domain)

    def refineMarkedElements(
        self,
        dim: int,
        elements: npt.NDArray[np.int32],
        netgen_flags: dict | None = None,
    ):
        """Refine mesh based on marked elements."""
        netgen_flags = netgen_flags or {}
        refine_faces = netgen_flags.get("refine_faces", False)
        gdim = self._mesh.geometry.dim
        if gdim not in (2, 3):
            raise RuntimeError("Refinement of 2D and 3D meshes is supported only.")
        # Gather all the element indices for refinement on rank 0.
        # Map elements to incidient cells.
        self._mesh.topology.create_connectivity(dim, self._mesh.topology.dim)
        local_cells = dolfinx.mesh.compute_incident_entities(
            self._mesh.topology, elements, dim, self._mesh.topology.dim
        )
        igi = self._mesh.topology.original_cell_index[local_cells]
        gathered_igi = self._mesh.comm.allgather(igi)

        ng_elements = _dim_to_element_wrapper(self.ngmesh)[self._mesh.topology.dim]()
        ng_dimension = len(ng_elements)
        marker = np.zeros(ng_dimension, dtype=np.int8)
        marker[self._sorted_mapping[np.hstack(gathered_igi)]] = 1
        for refine, el in zip(marker, ng_elements, strict=True):
            if refine:
                el.refine = True
            else:
                el.refine = False
        if not refine_faces and dim == 3:
            _dim_to_element_wrapper(self.ngmesh)[2]().Numpy()["refine"] = 0
        self.ngmesh.Refine(adaptive=True)
        self.ngmesh.Curve(1)  # Reset mesh to be linear

        self._mesh.comm.Barrier()
        ct, ft = self.extract_linear_mesh(gdim=gdim)
        return self._mesh, (ct, ft)


def extract_element_tags(
    comm_rank: int, ngmesh, dolfinx_mesh: dolfinx.mesh.Mesh, dim: int
) -> dolfinx.mesh.MeshTags:
    """
    Extract element tags from a Netgen mesh (on a given MPI rank) and distribute them onto
    the corresponding DOLFINx mesh.

    Args:
        comm_rank: The MPI rank to extract the element from.
        ngmesh: The Netgen mesh object.
        dolfinx_mesh: The DOLFINx mesh to which the facet tags will be distributed to.
        dim: The topological dimension of the entities to extract.
    """
    tdim = dolfinx_mesh.topology.dim
    assert ngmesh.dim == tdim, f"Mismatch: ({ngmesh.dim=}!={tdim=})"
    assert dolfinx_mesh.geometry.cmap.degree == 1, (
        "Can only extract element tags from linear grids"
    )
    comm = dolfinx_mesh.comm
    sub_entities = basix.cell.subentity_types(dolfinx_mesh.basix_cell())[dim]
    assert len(np.unique(sub_entities)) == 1, "Only one subentity type is supported"
    entity_type = dolfinx.mesh.to_type(sub_entities[0].name)
    num_vertices_per_cell = dolfinx.cpp.mesh.cell_num_vertices(entity_type)
    if comm.rank == comm_rank:
        ng_entities = _dim_to_element_wrapper(ngmesh)[dim]()
        element_indices = ng_entities.NumPy()["nodes"].astype(np.int64)
        entities = _numpy_trim(element_indices, dtype=np.int64) - 1
        if dim == dolfinx_mesh.topology.dim:
            entity_markers = ng_entities.NumPy()["index"].astype(np.int32)
        else:
            # Can't use the vectorized version, due to a bug in ngsolve:
            # https://forum.ngsolve.org/t/extract-facet-markers-from-netgen-mesh/3256
            entity_markers = np.array(
                [entity.index for entity in ng_entities], dtype=np.int32
            )
    else:
        # NOTE: Mixed meshes on non-simplex geometries requires changes
        entities = np.zeros((0, num_vertices_per_cell), dtype=np.int64)
        entity_markers = np.zeros((0,), dtype=np.int32)

    local_entities, local_values = dolfinx.io.distribute_entity_data(
        dolfinx_mesh, dim, entities, entity_markers
    )
    dolfinx_mesh.topology.create_connectivity(dim, 0)
    adj = dolfinx.graph.adjacencylist(local_entities)
    meshtag = dolfinx.mesh.meshtags_from_entities(
        dolfinx_mesh,
        dim,
        adj,
        local_values,
    )
    return meshtag



'''
This module contains all the functions related
'''
try:
    import firedrake as fd
    from firedrake.cython import mgimpl as impl, dmcommon
    from firedrake.__future__ import interpolate
    from firedrake import dmhooks
    import ufl
except ImportError:
    fd = None

from fractions import Fraction
import numpy as np
from petsc4py import PETSc

from netgen.meshing import MeshingParameters

from ngsPETSc.utils.firedrake.meshes import flagsUtils

def snapToNetgenDMPlex(ngmesh, petscPlex):
    '''
    This function snaps the coordinates of a DMPlex mesh to the coordinates of a Netgen mesh.
    '''
    if petscPlex.getDimension() == 2:
        ngCoordinates = ngmesh.Coordinates()
        petscCoordinates = petscPlex.getCoordinatesLocal().getArray().reshape(-1, ngmesh.dim)
        for i, pt in enumerate(petscCoordinates):
            j = np.argmin(np.sum((ngCoordinates - pt)**2, axis=1))
            petscCoordinates[i] = ngCoordinates[j]
        petscPlexCoordinates = petscPlex.getCoordinatesLocal()
        petscPlexCoordinates.setArray(petscPlexCoordinates)
        petscPlex.setCoordinatesLocal(petscPlexCoordinates)
    else:
        raise NotImplementedError("Snapping to Netgen meshes is only implemented for 2D meshes.")

def snapToCoarse(coarse, linear, degree, snap_smoothing, cg):
    '''
    This function snaps the coordinates of a DMPlex mesh to the coordinates of a Netgen mesh.
    '''
    dim = linear.geometric_dimension()
    if dim == 2:
        space = fd.VectorFunctionSpace(linear, "CG", degree)
        ho = fd.assemble(interpolate(coarse, space))
        if snap_smoothing == "hyperelastic":
            #Hyperelastic Smoothing
            bcs = [fd.DirichletBC(space, ho, "on_boundary")]
            quad_degree = 2*(degree+1)-1
            d = linear.topological_dimension()
            Q = fd.TensorFunctionSpace(linear, "DG", degree=0)
            Jinv = ufl.JacobianInverse(linear)
            hinv = fd.Function(Q)
            hinv.interpolate(Jinv)
            G = ufl.Jacobian(linear) * hinv
            ijac = 1/abs(ufl.det(G))

            def ref_grad(u):
                return ufl.dot(ufl.grad(u),G)

            params = {
                "snes_type": "newtonls",
                "snes_linesearch_type": "l2",
                "snes_max_it": 50,
                "snes_rtol": 1E-8,
                "snes_atol": 1E-8,
                "snes_ksp_ew": True,
                "snes_ksp_ew_rtol0": 1E-2,
                "snes_ksp_ew_rtol_max": 1E-2,
            }
            params["mat_type"] = "aij"
            coarse = {
                "ksp_type": "preonly",
                "pc_type": "lu",
                "pc_mat_factor_type": "mumps",
            }
            gmg = {
                "pc_type": "mg",
                "mg_coarse": coarse,
                "mg_levels": {
                    "ksp_max_it": 2,
                    "ksp_type": "chebyshev",
                    "pc_type": "jacobi",
                },
            }
            l = fd.mg.utils.get_level(linear)[1]
            pc = gmg if l else coarse
            params.update(pc)
            ksp = {
                "ksp_rtol": 1E-8,
                "ksp_atol": 0,
                "ksp_type": "minres",
                "ksp_norm_type": "preconditioned",
            }
            params.update(ksp)
            u = ho
            F = ref_grad(u)
            J = ufl.det(F)
            psi = (1/2) * (ufl.inner(F, F)-d - ufl.ln(J**2))
            U = (psi * ijac)*fd.dx(degree=quad_degree)
            dU = ufl.derivative(U, u, fd.TestFunction(space))
            problem = fd.NonlinearVariationalProblem(dU, u, bcs)
            solver = fd.NonlinearVariationalSolver(problem, solver_parameters=params)
            solver.set_transfer_manager(None)
            ctx = solver._ctx
            for c in problem.F.coefficients():
                dm = c.function_space().dm
                dmhooks.push_appctx(dm, ctx)
            solver.solve()
        if not cg:
            element = ho.function_space().ufl_element().sub_elements[0].reconstruct(degree=degree)
            space = fd.VectorFunctionSpace(linear, fd.BrokenElement(element))
            ho = fd.Function(space).interpolate(ho)
    else:
        raise NotImplementedError("Snapping to Netgen meshes is only implemented for 2D meshes.")
    return fd.Mesh(ho, comm=linear.comm, distribution_parameters=linear._distribution_parameters)

def uniformRefinementRoutine(ngmesh, cdm):
    '''
    Routing called inside of NetgenHierarchy to compute refined ngmesh and plex.
    '''
    #We refine the netgen mesh uniformly
    ngmesh.Refine(adaptive=False)
    #We refine the DMPlex mesh uniformly
    cdm.setRefinementUniform(True)
    rdm = cdm.refine()
    rdm.removeLabel("pyop2_core")
    rdm.removeLabel("pyop2_owned")
    rdm.removeLabel("pyop2_ghost")
    return (rdm, ngmesh)

def uniformMapRoutine(meshes, lgmaps):
    '''
    This function computes the coarse to fine and fine to coarse maps
    for a uniform mesh hierarchy.
    '''
    refinements_per_level = 1
    coarse_to_fine_cells = []
    fine_to_coarse_cells = [None]
    for (coarse, fine), (clgmaps, flgmaps) in zip(zip(meshes[:-1], meshes[1:]),
                                                zip(lgmaps[:-1], lgmaps[1:])):
        c2f, f2c = impl.coarse_to_fine_cells(coarse, fine, clgmaps, flgmaps)
        coarse_to_fine_cells.append(c2f)
        fine_to_coarse_cells.append(f2c)

    coarse_to_fine_cells = dict((Fraction(i, refinements_per_level), c2f)
                                for i, c2f in enumerate(coarse_to_fine_cells))
    fine_to_coarse_cells = dict((Fraction(i, refinements_per_level), f2c)
                                for i, f2c in enumerate(fine_to_coarse_cells))
    return (coarse_to_fine_cells, fine_to_coarse_cells)

def alfeldRefinementRoutine(ngmesh, cdm):
    '''
    Routing called inside of NetgenHierarchy to compute refined ngmesh and plex.
    '''
    #We refine the netgen mesh alfeld
    ngmesh.SplitAlfeld()
    #We refine the DMPlex mesh alfeld
    tr = PETSc.DMPlexTransform().create(comm=PETSc.COMM_WORLD)
    tr.setType(PETSc.DMPlexTransformType.REFINEREGULAR)
    tr.setDM(cdm)
    tr.setUp()
    rdm = tr.apply(cdm)
    return (rdm, ngmesh)

def alfeldMapRoutine(meshes):
    '''
    This function computes the coarse to fine and fine to coarse maps
    for a alfeld mesh hierarchy.
    '''
    raise NotImplementedError("Alfeld refinement is not implemented yet.")

refinementTypes = {"uniform": (uniformRefinementRoutine, uniformMapRoutine),
                   "Alfeld": (alfeldRefinementRoutine, alfeldMapRoutine)}

def NetgenHierarchy(mesh, levs, flags):
    '''
    This function creates a Firedrake mesh hierarchy from Netgen/NGSolve meshes.

    :arg mesh: the Netgen/NGSolve mesh
    :arg levs: the number of levels in the hierarchy
    :arg netgen_flags: either a bool or a dictionray containing options for Netgen.
    If not False the hierachy is constructed using ngsPETSc, if None hierarchy
    constructed in a standard manner. Netgen flags includes:
        -degree, either an integer denoting the degree of curvature of all levels of
        the mesh or a list of levs+1 integers denoting the degree of curvature of
        each level of the mesh.
        -tol, geometric tolerance adopted in snapToNetgenDMPlex.
        -refinement_type, the refinment type to be used: uniform (default), Alfeld
    '''
    if mesh.geometric_dimension() == 3:
        raise NotImplementedError("Netgen hierachies are only implemented for 2D meshes.")
    comm = mesh.comm
    #Parsing netgen flags
    if not isinstance(flags, dict):
        flags = {}
    order = flagsUtils(flags, "degree", 1)
    if isinstance(order, int):
        order= [order]*(levs+1)
    permutation_tol = flagsUtils(flags, "tol", 1e-8)
    refType = flagsUtils(flags, "refinement_type", "uniform")
    optMoves = flagsUtils(flags, "optimisation_moves", False)
    snap = flagsUtils(flags, "snap_to", "geometry")
    snap_smoothing = flagsUtils(flags, "snap_smoothing", "hyperelastic")
    cg = flagsUtils(flags, "cg", False)
    nested = flagsUtils(flags, "nested", snap in ["coarse"])
    #Firedrake quoantities
    meshes = []
    lgmaps = []
    params = {"partition": False}
    #We curve the mesh
    if order[0]>1:
        ho_field = mesh.curve_field(
            order=order[0],
            permutation_tol=permutation_tol,
            cg_field=cg
        )
        temp = fd.Mesh(ho_field,distribution_parameters=params, comm=comm)
        temp.netgen_mesh = mesh.netgen_mesh
        temp._tolerance = mesh.tolerance
        mesh = temp
    # Make a plex (cdm) without overlap.
    dm_cell_type, = mesh.dm_cell_types
    tdim = mesh.topology_dm.getDimension()
    cdm = dmcommon.submesh_create(mesh.topology_dm, tdim, "celltype", dm_cell_type, True)
    cdm.removeLabel("pyop2_core")
    cdm.removeLabel("pyop2_owned")
    cdm.removeLabel("pyop2_ghost")
    no = impl.create_lgmap(cdm)
    o = impl.create_lgmap(mesh.topology_dm)
    lgmaps.append((no, o))
    mesh.topology_dm.setRefineLevel(0)
    meshes += [mesh]
    ngmesh = mesh.netgen_mesh
    for l in range(levs):
        #Streightening the mesh
        ngmesh.Curve(1)
        rdm, ngmesh = refinementTypes[refType][0](ngmesh, cdm)
        cdm = rdm
        #We snap the mesh to the Netgen mesh
        if snap == "geometry":
            snapToNetgenDMPlex(ngmesh, rdm)
        #We construct a Firedrake mesh from the DMPlex mesh
        no = impl.create_lgmap(rdm)
        mesh = fd.Mesh(rdm, dim=meshes[-1].geometric_dimension(), reorder=False,
                       distribution_parameters=params, comm=comm)
        o = impl.create_lgmap(mesh.topology_dm)
        lgmaps.append((no, o))
        if optMoves:
            #Optimises the mesh, for example smoothing
            if ngmesh.dim == 2:
                ngmesh.OptimizeMesh2d(MeshingParameters(optimize2d=optMoves))
            elif mesh.dim == 3:
                ngmesh.OptimizeVolumeMesh(MeshingParameters(optimize3d=optMoves))
            else:
                raise ValueError("Only 2D and 3D meshes can be optimised.")
        mesh.netgen_mesh = ngmesh
        #We curve the mesh
        if order[l+1] > 1:
            if snap == "geometry":
                mesh = fd.Mesh(
                    mesh.curve_field(order=order[l+1], permutation_tol=permutation_tol),
                    distribution_parameters=params,
                    comm=comm
                )
            elif snap == "coarse":
                mesh = snapToCoarse(ho_field, mesh, order[l+1], snap_smoothing, cg)
        mesh.topology_dm.setRefineLevel(1 + l)
        meshes += [mesh]
    #We populate the coarse to fine map
    coarse_to_fine_cells, fine_to_coarse_cells = refinementTypes[refType][1](meshes, lgmaps)
    return fd.HierarchyBase(meshes, coarse_to_fine_cells, fine_to_coarse_cells,
                            1, nested=nested)



'''
This module contains all the function and class needed to wrap a PETSc Nullspace in NGSolve
'''
from petsc4py import PETSc
from ngsPETSc import VectorMapping

class NullSpace:
    '''
    This class creates a PETSc Null space from NGSolve vectors

    :arg fes: NGSolve finite element space.
    :arg span: list NGSolve grid functions spanning the null space.

    '''
    def __init__(self, fes, span, near=False):
        constant = False
        self.vecMap = VectorMapping(fes)
        self.near = near
        if isinstance(span, list):
            nullspace = list(span)
            petscNullspace = []
            for vec in nullspace:
                if isinstance(vec, str):
                    if vec == "constant":
                        constant = True
                    else:
                        raise ValueError("Invalid nullspace string")
                else:
                    petscVec  = self.vecMap.petscVec(vec)
                    petscNullspace.append(petscVec)
        elif isinstance(span, str):
            if span == "constant":
                constant = True
                petscNullspace = []
            else:
                raise ValueError("Invalid nullspace string")
        else:
            raise ValueError("Invalid nullspace type")
            # Create vector space basis and orthogonalize
        self.constant = constant
        self.orthonormalize(petscNullspace)
        self.nullspace = PETSc.NullSpace().create(constant=constant, vectors=petscNullspace)
    def orthonormalize(self, basis):
        """Orthonormalize the basis."""
        for i, vec in enumerate(basis):
            alphas = []
            for vec_ in basis[:i]:
                alphas.append(vec.dot(vec_))
            for alpha, vec_ in zip(alphas, basis[:i]):
                vec.axpy(-alpha, vec_)
            if self.constant:
                # Subtract constant mode
                alpha = vec.sum()
                vec.array -= alpha
            vec.normalize()



'''
This module contains all the functions related to wrapping NGSolve meshes to Firedrake
We adopt the same docstring conventions as the Firedrake project, since this part of
the package will only be used in combination with Firedrake.
'''
try:
    import firedrake as fd
    from firedrake.__future__ import interpolate
except ImportError:
    fd = None

import numpy as np
from petsc4py import PETSc

import netgen
import netgen.meshing as ngm
from netgen.meshing import MeshingParameters
try:
    import ngsolve as ngs
except ImportError:
    class ngs:
        "dummy class"
        class comp:
            "dummy class"
            Mesh = type(None)

from ngsPETSc import MeshMapping
from ngsPETSc.utils.utils import find_permutation

def flagsUtils(flags, option, default):
    '''
    utility fuction used to parse Netgen flag options
    '''
    try:
        return flags[option]
    except KeyError:
        return default

def refineMarkedElements(self, mark, netgen_flags={}):
    '''
    This method is used to refine a mesh based on a marking function
    which is a Firedrake DG0 function.

    :arg mark: the marking function which is a Firedrake DG0 function.
    :arg netgen_flags: the dictionary of flags to be passed to ngsPETSc.
    It includes the option:
        - refine_faces, which is a boolean specifyiong if you want to refine faces.

    '''
    DistParams = self._distribution_parameters
    els = {2: self.netgen_mesh.Elements2D, 3: self.netgen_mesh.Elements3D}
    dim = self.geometric_dimension()
    refine_faces = flagsUtils(netgen_flags, "refine_faces", False)
    if dim in [2,3]:
        with mark.dat.vec as marked:
            marked0 = marked
            getIdx = self._cell_numbering.getOffset
            if self.sfBCInv is not None:
                getIdx = lambda x: x #pylint: disable=C3001
                _, marked0 = self.topology_dm.distributeField(self.sfBCInv,
                                                              self._cell_numbering,
                                                              marked)
            if self.comm.Get_rank() == 0:
                mark = marked0.getArray()
                max_refs = np.max(mark)
                for _ in range(int(max_refs)):
                    for i, el in enumerate(els[dim]()):
                        if mark[getIdx(i)] > 0:
                            el.refine = True
                        else:
                            el.refine = False
                    if not refine_faces and dim == 3:
                        self.netgen_mesh.Elements2D().NumPy()["refine"] = 0
                    self.netgen_mesh.Refine(adaptive=True)
                    mark = mark-np.ones(mark.shape)
                return fd.Mesh(self.netgen_mesh, distribution_parameters=DistParams, comm=self.comm)
            return fd.Mesh(netgen.libngpy._meshing.Mesh(dim),
                           distribution_parameters=DistParams, comm=self.comm)
    else:
        raise NotImplementedError("No implementation for dimension other than 2 and 3.")



@PETSc.Log.EventDecorator()
def curveField(self, order, permutation_tol=1e-8, location_tol=1e-1, cg_field=False):
    '''
    This method returns a curved mesh as a Firedrake function.

    :arg order: the order of the curved mesh.
    :arg permutation_tol: tolerance used to construct the permutation of the reference element.
    :arg location_tol: tolerance used to locate the cell a point belongs to.
    :arg cg_field: return a CG function field representing the mesh, rather than the
                   default DG field.

    '''
    # Check if the mesh is a surface mesh or two dimensional mesh
    if len(self.netgen_mesh.Elements3D()) == 0:
        ng_element = self.netgen_mesh.Elements2D
    else:
        ng_element = self.netgen_mesh.Elements3D
    ng_dimension = len(ng_element())
    geom_dim = self.geometric_dimension()

    # Construct the mesh as a Firedrake function
    if cg_field:
        firedrake_space = fd.VectorFunctionSpace(self, "CG", order)
    else:
        low_order_element = self.coordinates.function_space().ufl_element().sub_elements[0]
        ufl_element = low_order_element.reconstruct(degree=order)
        firedrake_space = fd.VectorFunctionSpace(self, fd.BrokenElement(ufl_element))
    new_coordinates = fd.assemble(interpolate(self.coordinates, firedrake_space))

    # Compute reference points using fiat
    fiat_element = new_coordinates.function_space().finat_element.fiat_equivalent
    entity_ids = fiat_element.entity_dofs()
    nodes = fiat_element.dual_basis()
    ref = []
    for dim in entity_ids:
        for entity in entity_ids[dim]:
            for dof in entity_ids[dim][entity]:
                # Assert singleton point for each node.
                pt, = nodes[dof].get_point_dict().keys()
                ref.append(pt)
    reference_space_points = np.array(ref)

    # Curve the mesh on rank 0 only
    if self.comm.rank == 0:
        # Construct numpy arrays for physical domain data
        physical_space_points = np.zeros(
            (ng_dimension, reference_space_points.shape[0], geom_dim)
        )
        curved_space_points = np.zeros(
            (ng_dimension, reference_space_points.shape[0], geom_dim)
        )
        self.netgen_mesh.CalcElementMapping(reference_space_points, physical_space_points)
        self.netgen_mesh.Curve(order)
        self.netgen_mesh.CalcElementMapping(reference_space_points, curved_space_points)
        curved = ng_element().NumPy()["curved"]
        # Broadcast a boolean array identifying curved cells
        curved = self.comm.bcast(curved, root=0)
        physical_space_points = physical_space_points[curved]
        curved_space_points = curved_space_points[curved]
    else:
        curved = self.comm.bcast(None, root=0)
        # Construct numpy arrays as buffers to receive physical domain data
        ncurved = np.sum(curved)
        physical_space_points = np.zeros(
            (ncurved, reference_space_points.shape[0], geom_dim)
        )
        curved_space_points = np.zeros(
            (ncurved, reference_space_points.shape[0], geom_dim)
        )

    # Broadcast curved cell point data
    self.comm.Bcast(physical_space_points, root=0)
    self.comm.Bcast(curved_space_points, root=0)
    cell_node_map = new_coordinates.cell_node_map()

    # Select only the points in curved cells
    barycentres = np.average(physical_space_points, axis=1)
    ng_index = [*map(lambda x: self.locate_cell(x, tolerance=location_tol), barycentres)]

    # Select only the indices of points owned by this rank
    owned = [(0 <= ii < len(cell_node_map.values)) if ii is not None else False for ii in ng_index]

    # Select only the points owned by this rank
    physical_space_points = physical_space_points[owned]
    curved_space_points = curved_space_points[owned]
    barycentres = barycentres[owned]
    ng_index = [idx for idx, o in zip(ng_index, owned) if o]

    # Get the PyOP2 indices corresponding to the netgen indices
    pyop2_index = []
    for ngidx in ng_index:
        pyop2_index.extend(cell_node_map.values[ngidx])

    # Find the correct coordinate permutation for each cell
    # NB: Coordinates must be cast to real when running Firedrake in complex mode
    permutation = find_permutation(
        physical_space_points,
        new_coordinates.dat.data[pyop2_index].reshape(
            physical_space_points.shape
        ).astype(np.float64, copy=False),
        tol=permutation_tol
    )

    # Apply the permutation to each cell in turn
    for ii, p in enumerate(curved_space_points):
        curved_space_points[ii] = p[permutation[ii]]

    # Assign the curved coordinates to the dat
    new_coordinates.dat.data[pyop2_index] = curved_space_points.reshape(-1, geom_dim)

    return new_coordinates

def splitToQuads(plex, dim, comm):
    '''
    This method splits a Netgen mesh to quads, using a PETSc transform.
    TODO: Improve support quad meshing.
        @pef  Get netgen to make a quad-dominant mesh, and then only split the triangles.
              Current implementation will make for poor-quality meshes.
    '''
    if dim == 2:
        transform = PETSc.DMPlexTransform().create(comm=comm)
        transform.setType(PETSc.DMPlexTransformType.REFINETOBOX)
        transform.setDM(plex)
        transform.setUp()
    else:
        raise RuntimeError("Splitting to quads is only possible for 2D meshes.")
    newplex = transform.apply(plex)
    return newplex

splitTypes = {"Alfeld": lambda x: x.SplitAlfeld(),
              "Powell-Sabin": lambda x: x.SplitPowellSabin()}

class FiredrakeMesh:
    '''
    This class creates a Firedrake mesh from Netgen/NGSolve meshes.

    :arg mesh: the mesh object, it can be either a Netgen/NGSolve mesh or a PETSc DMPlex
    :param netgen_flags: The dictionary of flags to be passed to ngsPETSc.
    :arg comm: the MPI communicator.
    '''
    def __init__(self, mesh, netgen_flags, user_comm=fd.COMM_WORLD):
        self.comm = user_comm
        #Parsing netgen flags
        if not isinstance(netgen_flags, dict):
            netgen_flags = {}
        split2tets = flagsUtils(netgen_flags, "split_to_tets", False)
        split = flagsUtils(netgen_flags, "split", False)
        quad = flagsUtils(netgen_flags, "quad", False)
        optMoves = flagsUtils(netgen_flags, "optimisation_moves", False)
        #Checking the mesh format
        if isinstance(mesh,(ngs.comp.Mesh,ngm.Mesh)):
            if split2tets:
                mesh = mesh.Split2Tets()
            if split:
                #Split mesh this includes Alfeld and Powell-Sabin
                splitTypes[split](mesh)
            if optMoves:
                #Optimises the mesh, for example smoothing
                if mesh.dim == 2:
                    mesh.OptimizeMesh2d(MeshingParameters(optimize2d=optMoves))
                elif mesh.dim == 3:
                    mesh.OptimizeVolumeMesh(MeshingParameters(optimize3d=optMoves))
                else:
                    raise ValueError("Only 2D and 3D meshes can be optimised.")
            #We create the plex from the netgen mesh
            self.meshMap = MeshMapping(mesh, comm=self.comm)
            #We apply the DMPLEX transform
            if quad:
                newplex = splitToQuads(self.meshMap.petscPlex, mesh.dim, comm=self.comm)
                self.meshMap = MeshMapping(newplex)
        elif isinstance(mesh, PETSc.DMPlex):
            self.meshMap = MeshMapping(mesh)
        else:
            raise ValueError("Mesh format not recognised.")

    def createFromTopology(self, topology, name, comm):
        '''
        Internal method to construct a mesh from a mesh topology, copied from Firedrake.

        :arg topology: the mesh topology

        :arg name: the mesh name

        '''
        cell = topology.ufl_cell()
        geometric_dim = topology.topology_dm.getCoordinateDim()
        element = fd.VectorElement("Lagrange", cell, 1, dim=geometric_dim)
        # Create mesh object
        self.firedrakeMesh = fd.MeshGeometry.__new__(fd.MeshGeometry, element, comm)
        self.firedrakeMesh._init_topology(topology)
        self.firedrakeMesh.name = name
        # Adding Netgen mesh and inverse sfBC as attributes
        self.firedrakeMesh.netgen_mesh = self.meshMap.ngMesh
        if self.firedrakeMesh.sfBC is not None:
            self.firedrakeMesh.sfBCInv = self.firedrakeMesh.sfBC.createInverse()
        else:
            self.firedrakeMesh.sfBCInv = None
        #Generating ngs to Firedrake cell index map
        #Adding refine_marked_elements and curve_field methods
        setattr(fd.MeshGeometry, "refine_marked_elements", refineMarkedElements)
        setattr(fd.MeshGeometry, "curve_field", curveField)



'''
This module contains all the functions related to the PETSc linear
system solver (KSP) interface for NGSolve
'''
from petsc4py import PETSc
from ngsolve import la, BilinearForm, FESpace, BitArray, Projector
from ngsPETSc import Matrix, VectorMapping, PETScPreconditioner, NullSpace

def createFromBilinearForm(a, freeDofs, solverParameters):
    """
    This function creates a PETSc matrix from an NGSolve bilinear form
    """
    a.Assemble()
    #Setting deafult matrix type
    if "ngs_mat_type" not in solverParameters:
        solverParameters["ngs_mat_type"] = "aij"
    #Assembling matrix if not of type Python
    if solverParameters["ngs_mat_type"] not in ["python"]:
        if hasattr(a.mat, "row_pardofs"):
            dofs = a.mat.row_pardofs
        else:
            dofs = None
        mat = Matrix(a.mat, (dofs, freeDofs, None), solverParameters["ngs_mat_type"])
        return (a.mat, mat.mat)
    raise ValueError("ngs_mat_type {} is not supported.".format(solverParameters["ngs_mat_type"]))

def createFromMatrix(a, freeDofs, solverParameters):
    """
    This function creates a PETSc matrix from an NGSolve bilinear form
    """
    #Setting deafult matrix type
    if "ngs_mat_type" not in solverParameters:
        solverParameters["ngs_mat_type"] = "aij"
    #Assembling matrix if not of type Python
    if solverParameters["ngs_mat_type"] not in ["python"]:
        if hasattr(a, "row_pardofs"):
            dofs = a.row_pardofs
        else:
            dofs = None
        mat = Matrix(a, (dofs, freeDofs, None), solverParameters["ngs_mat_type"])
        pscMat = mat.mat
        return (a, pscMat)
    if solverParameters["ngs_mat_type"] == "python":
        _, pscMat = createFromAction(a, freeDofs, solverParameters)
        return (a, pscMat)
    raise ValueError("ngs_mat_type {} is not supported.".format(solverParameters["ngs_mat_type"]))

def createFromPC(a, freeDofs, solverParameters):
    """
    This function creates a PETSc matrix from an ngsPETSc PETSc Preconditioner
    """
    class Wrap(object):
        """
        This class wraps a PETSc Preconditioner as PETSc Python matrix
        """
        def __init__(self, a, dofs, freeDofs):
            self.a = a
            self.mapping = VectorMapping((dofs,freeDofs,{"bsize": [1]}))
            self.ngX = a.CreateColVector()
            self.ngY = a.CreateColVector()
            if hasattr(a, "actingDofs"):
                self.prj = Projector(mask=a.actingDofs, range=True)
            else:
                self.prj = Projector(mask=freeDofs, range=True)

        def mult(self, mat, X, Y): #pylint: disable=W0613
            """
            PETSc matrix-vector product
            """
            self.mapping.ngsVec(X, (self.prj*self.ngX).Evaluate())
            self.a.Mult(self.ngX, self.ngY)
            self.mapping.petscVec((self.prj*self.ngY).Evaluate(), Y)
    #Grabbing comm information
    if hasattr(a, "row_pardofs"):
        dofs = a.row_pardofs
        comm = dofs.comm.mpi4py
    elif "dofs" in solverParameters:
        dofs = solverParameters["dofs"]
        comm = dofs.comm.mpi4py
    else:
        dofs = None
        comm = PETSc.COMM_SELF
    pythonA = Wrap(a, dofs, freeDofs)
    pscA = PETSc.Mat().create(comm=comm)
    pscA.setSizes([sum(freeDofs), sum(freeDofs)])
    pscA.setType("python")
    pscA.setPythonContext(pythonA)
    pscA.setUp()
    return (a.ngsMat, pscA)

def createFromAction(a, freeDofs, solverParameters):
    """
    This function creates a matrix free PETSc matrix from an NGSolve matrix
    """
    class Wrap(object):
        """
        This class wraps an NGSolve matrix as PETSc Python matrix
        """
        def __init__(self, a, dofs, freeDofs, comm):
            self.a = a
            self.dofs = dofs
            self.freeDofs = freeDofs
            self.comm = comm
            self.mapping = VectorMapping((dofs,freeDofs,{"bsize": (1,1)}))
            self.ngX = a.CreateColVector()
            self.ngY = a.CreateColVector()

        def mult(self, mat, X, Y): #pylint: disable=W0613
            """
            PETSc matrix-vector product
            """
            self.mapping.ngsVec(X, self.ngX)
            self.a.Mult(self.ngX, self.ngY)
            self.mapping.petscVec(self.ngY, Y)

    if hasattr(a, "row_pardofs"):
        dofs = a.row_pardofs
        comm = dofs.comm.mpi4py
        entrysize = a.local_mat.entrysizes[0]
        _, rnumberGlobal = dofs.EnumerateGlobally(freeDofs) #samrc
    elif "dofs" in solverParameters:
        dofs = solverParameters["dofs"]
        comm = dofs.comm.mpi4py
        entrysize = dofs.entrysize
        _, rnumberGlobal = dofs.EnumerateGlobally(freeDofs) #samrc
    else:
        dofs = None
        comm = PETSc.COMM_SELF
        entrysize = 1
        rnumberGlobal = sum(freeDofs)
    pythonA = Wrap(a, dofs, freeDofs, comm)
    pscA = PETSc.Mat().create(comm=comm)
    pscA.setSizes(size=(rnumberGlobal*entrysize,
                        rnumberGlobal*entrysize), bsize=entrysize)
    pscA.setType("python")
    pscA.setPythonContext(pythonA)
    pscA.setUp()
    return (a, pscA)

parse = {BilinearForm: createFromBilinearForm,
         la.SparseMatrixd: createFromMatrix,
         la.ParallelMatrix: createFromMatrix,
         PETScPreconditioner: createFromPC,
         la.BaseMatrix: createFromAction}

class KrylovSolver():
    """
    This class creates a PETSc Krylov Solver (KSP) for NGSolve.
    Inspired by Firedrake linear solver class.

    :arg a: either the bilinear form, ngs Matrix or a petsc4py matrix

    :arg dofsDescr: either finite element space

    :arg p: either the bilinear form, ngs Matrix or petsc4py matrix actin as a preconditioner

    :arg nullspace: either a PETSc NullSpace or ngsPETSc PETSc Preconditioner
    or touple of ngsPETSc PETSc Preconditioner.

    :arg solverParameters: parameters to be passed to the KS P solver

    :arg optionsPrefix: special solver options prefix for this specific Krylov solver

    """
    def __init__(self, a, dofsDescr, p=None, nullspace=None, optionsPrefix="",
                 solverParameters={}):
        # Grabbing dofs information
        if isinstance(dofsDescr, FESpace):
            freeDofs = dofsDescr.FreeDofs()
        elif isinstance(dofsDescr, BitArray):
            freeDofs = dofsDescr
        else:
            raise ValueError("dofsDescr must be either FESpace or BitArray")
        #Construct operator
        pscA = None
        for key in parse: #pylint: disable=C0206
            if isinstance(a, key):
                ngsA, pscA = parse[key](a, freeDofs, solverParameters)
        if pscA is None:
            raise ValueError("a of type {} not supported.".format(type(a)))
        if p is not None:
            for key in parse:  #pylint: disable=C0206
                if isinstance(p, key):
                    if hasattr(ngsA, "row_pardofs"):
                        solverParameters["dofs"] = ngsA.row_pardofs
                    _, pscP = parse[key](p, freeDofs, solverParameters)
                    break
        else:
            pscP = pscA
        #Construct vector mapping
        if hasattr(ngsA, "row_pardofs"):
            dofs = ngsA.row_pardofs
        else:
            dofs = None
        if hasattr(ngsA.local_mat, "entrysizes"):
            entrysize = ngsA.local_mat.entrysizes
        else:
            entrysize = [1]
        self.mapping = VectorMapping((dofs,freeDofs,{"bsize":entrysize}))
        #Fixing PETSc options
        options_object = PETSc.Options()
        for optName, optValue in solverParameters.items():
            options_object[optName] = optValue

        #Setting PETSc Options
        pscA.setOptionsPrefix(optionsPrefix)
        pscA.setFromOptions()

        #Setting up nullspace
        if nullspace is not None:
            if isinstance(nullspace, (list, tuple)):
                for ns in nullspace:
                    if isinstance(ns, PETSc.NullSpace):
                        pscA.setNullSpace(ns)
                    elif isinstance(ns, NullSpace):
                        if ns.near:
                            pscA.setNearNullSpace(ns.nullspace)
                        else:
                            pscA.setNullSpace(ns.nullspace)
                    else:
                        raise ValueError("nullspace must be either \
                                         PETSc.NullSpace or NullSpace")
            if isinstance(nullspace, PETSc.NullSpace):
                pscA.setNullSpace(nullspace)
            elif isinstance(nullspace, NullSpace):
                if nullspace.near:
                    pscA.setNearNullSpace(nullspace.nullspace)
                else:
                    pscA.setNullSpace(nullspace.nullspace)
            else:
                raise ValueError("nullspace must be either \
                                    PETSc.NullSpace or NullSpace")

        #Setting up KSP
        self.ksp = PETSc.KSP().create(comm=pscA.getComm())
        self.ksp.setOperators(A=pscA, P=pscP)
        self.ksp.setOptionsPrefix(optionsPrefix)
        self.ksp.setFromOptions()
        self.pscX, self.pscB = pscA.createVecs()

        #Attaching operator
        self.ngsA = ngsA

    def solve(self, b, x, mapping=None):
        """
        This function solves the linear system

        :arg b: right hand side of the linear system
        :arg x: solution of the linear system
        """
        if mapping is None:
            mapping = self.mapping
        mapping.petscVec(x, self.pscX)
        mapping.petscVec(b, self.pscB)
        self.ksp.solve(self.pscB, self.pscX)
        mapping.ngsVec(self.pscX, x)

    def operator(self):
        """
        This function returns the operator of the KSP solver
        """
        return KSPOpeator(self)

class KSPOpeator(la.BaseMatrix):
    """
    This class wraps a PETSc KSP solver as an NGSolve matrix
    """
    def __init__(self, ksp):
        la.BaseMatrix.__init__(self)
        self.ksp = ksp

    def Shape(self):
        '''
        Shape of the BaseMatrix

        '''
        return self.ksp.ngsA.shape

    def CreateVector(self,col):
        '''
        Create vector corresponding to the matrix

        :arg col: True if one want a column vector

        '''
        return self.ksp.ngsA.CreateVector(not col)

    def Mult(self, x, y):
        """
        Matrix-vector product
        """
        self.ksp.solve(x, y)



'''
This module contains all the functions related to wrapping NGSolve matrices to
PETSc matrices using the petsc4py interface.
'''
import numpy as np

from ngsolve import FESpace
from petsc4py import PETSc

class Matrix(object):
    '''
    This class creates a PETSc Matrix

    :arg ngsMat: the NGSolve matrix

    :arg freeDofs: free DOFs of the FE spaces used to construct the matrix

    :arg matType: type of PETSc matrix, i.e. PETSc sparse: aij,
    MKL sparse: aijmkl or CUDA: aijcusparse

    '''
    def __init__(self, ngsMat, parDescr, matType="aij", petscMat=None):
        if not isinstance(parDescr, (tuple, list)):
            samerc = True
            if isinstance(parDescr, FESpace):
                dofs = parDescr.ParallelDofs()
                self.freeDofs = (parDescr.FreeDofs(),parDescr.FreeDofs())
                comm = dofs.comm.mpi4py
            else:
                dofs, freeDofs, dofsInfo = parDescr
                self.freeDofs = (freeDofs,freeDofs)
                if dofs is not None:
                    comm = dofs.comm.mpi4py
                else:
                    ### create suitable dofs
                    comm = PETSc.COMM_SELF
                    dofs = type('', (object,), {'entrysize':dofsInfo["bsize"][0]})()
        else:
            samerc = False
            if isinstance(parDescr[0], FESpace) and isinstance(parDescr[1], FESpace):
                dofs = parDescr[0].ParallelDofs()
                self.freeDofs = (parDescr[0].FreeDofs(),parDescr[1].FreeDofs())
                comm = dofs.comm.mpi4py
            elif isinstance(parDescr[0], (tuple, list)):
                dofs = parDescr[0][0]
                self.freeDofs = (parDescr[0][1], parDescr[1][1])
                if dofs is not None:
                    comm = dofs.comm.mpi4py
                else:
                    ### create suitable dofs
                    comm = PETSc.COMM_SELF
            else:
                samerc = True
                dofs = parDescr[0]
                self.freeDofs = (parDescr[1], parDescr[1])
                if dofs is not None:
                    comm = dofs.comm.mpi4py
                else:
                    ### create suitable dofs
                    comm = PETSc.COMM_SELF

        localMat = ngsMat.local_mat
        entryHeight, entryWidth = localMat.entrysizes
        if entryHeight != entryWidth: raise RuntimeError ("Only square entries are allowed.")

        valMat, colMat, indMat = localMat.CSR()
        indMat = np.array(indMat).astype(PETSc.IntType)
        colMat = np.array(colMat).astype(PETSc.IntType)
        if entryHeight > 1:
            petscLocalMat = PETSc.Mat().createBAIJ(size=(entryHeight*localMat.height,
                                                         entryWidth*localMat.width),
                                                   bsize=entryHeight,
                                                   csr=(indMat,colMat,valMat),
                                                   comm=PETSc.COMM_SELF)
        else:
            petscLocalMat = PETSc.Mat().createAIJ(size=(localMat.height,
                                                        localMat.width),
                                                  csr=(indMat,colMat,valMat),
                                                  comm=PETSc.COMM_SELF)
        if self.freeDofs[0] is not None or self.freeDofs[1] is not None:
            rowIsFreeLocal = None
            if self.freeDofs[0] is not None:
                rowLocalMatFree = np.flatnonzero(self.freeDofs[0]).astype(PETSc.IntType)
                rowIsFreeLocal = PETSc.IS().createBlock(indices=rowLocalMatFree,
                                                        bsize=entryHeight,comm=PETSc.COMM_SELF)
            colIsFreeLocal = rowIsFreeLocal
            if self.freeDofs[1] is not None and not samerc:
                colLocalMatFree = np.flatnonzero(self.freeDofs[1]).astype(PETSc.IntType)
                colIsFreeLocal = PETSc.IS().createBlock(indices=colLocalMatFree,
                                                        bsize=entryHeight,comm=PETSc.COMM_SELF)
            petscLocalMat = petscLocalMat.createSubMatrices(rowIsFreeLocal, colIsFreeLocal)[0]

        if comm.Get_size() > 1:
            # is this a BUG in the bindings?
            #rparallelDofs = ngsMat.row_pardofs
            rparallelDofs = ngsMat.col_pardofs
            rglobalNums, rnumberGlobal = rparallelDofs.EnumerateGlobally(self.freeDofs[0])
            if self.freeDofs[0] is not None:
                rglobalNums = np.array(rglobalNums, dtype=PETSc.IntType)[self.freeDofs[0]]
            rlocalGlobalMap = PETSc.LGMap().create(indices=rglobalNums,
                                                   bsize=entryHeight,
                                                   comm=comm)
            if not samerc:
                # is this a BUG in the bindings?
                #cparallelDofs = ngsMat.col_pardofs
                cparallelDofs = ngsMat.row_pardofs
                cglobalNums, cnumberGlobal = cparallelDofs.EnumerateGlobally(self.freeDofs[1])
                if self.freeDofs[1] is not None:
                    cglobalNums = np.array(cglobalNums, dtype=PETSc.IntType)[self.freeDofs[1]]
                clocalGlobalMap = PETSc.LGMap().create(indices=cglobalNums,
                                                       bsize=entryWidth,
                                                       comm=comm)
            else:
                clocalGlobalMap = rlocalGlobalMap
                cnumberGlobal = rnumberGlobal

            mat = PETSc.Mat().create(comm=comm)
            mat.setSizes(size=(rnumberGlobal*entryHeight,
                            cnumberGlobal*entryHeight), bsize=entryHeight)
            mat.setType(PETSc.Mat.Type.IS)
            mat.setLGMap(rlocalGlobalMap, clocalGlobalMap)
            mat.setISLocalMat(petscLocalMat)
            mat.assemble()
            if matType != 'is':
                mat.convert(matType)
            if petscMat is None:
                self.mat = mat
            else:
                mat.copy(petscMat)
        else:
            mat = petscLocalMat
            mat.convert(matType)
            if matType != 'is':
                mat.convert(matType)

            if petscMat is None:
                self.mat = mat
            else:
                mat.copy(petscMat)

    def view(self):
        '''
        This function display PETSc Mat info

        '''
        self.mat.view()



'''
This module contains all the functions related to the PETSc SNES
'''
from petsc4py import PETSc

from ngsolve import GridFunction

from ngsPETSc import VectorMapping, Matrix


class NonLinearSolver:
    '''
    This class creates a PETSc Non-Linear Solver (SNES) from a callback to
    a NGSolve residual vector

    :arg fes: the finite element space over which the non-linear problem
              is defined

    :arg a: the variational form reppresenting the non-linear problem

    :arg residual: callback to the residual for the non-linear solver,
                   this fuction is used only if the argument a is None.

    :arg objective: callback to the objective for the non-linear solver,
                   this fuction is used only if the argument a is None,
                   if False the PETSSc default is norm 2.

    :arg jacobian: callback to the Jacobian for the non-linear solver,
                   this fuction is used only if the argument a is None.
    '''
    def __init__(self, fes, a=None, residual=None, objective=None, jacobian=None,
                 solverParameters={}, optionsPrefix=""):
        self.fes = fes
        dofs = fes.ParallelDofs()
        self.second_order = False
        if "ngs_jacobian_mat_type" not in solverParameters:
            solverParameters["ngs_jacobian_mat_type"] = "aij"
        jacobianMatType = solverParameters["ngs_jacobian_mat_type"]
        self.snes = PETSc.SNES().create(comm=dofs.comm.mpi4py)
        #Setting up the options
        options_object = PETSc.Options()
        for optName, optValue in solverParameters.items():
            options_object[optName] = optValue
        self.snes.setOptionsPrefix(optionsPrefix)
        self.snes.setFromOptions()
        #Setting up utility for mappings
        self.vectorMapping = VectorMapping(self.fes)
        if residual is not None: self.residual = residual
        elif a is not None:
            def residual(x):  #pylint: disable=E0102, E0213
                res = GridFunction(fes)
                a.Apply(x.vec, res.vec)
                return res
            self.residual = residual
        else:
            raise ValueError("Either evalFunction or a must be provided")
        if objective is not None: self.objective = objective
        elif a is not None:
            def objective(x):  #pylint: disable=E0102, E0213
                return a.Energy(x.vec)
            self.objective = objective
        if jacobian is not None:
            self.jacobian = jacobian
            self.jacobianMatType = jacobianMatType
            self.second_order = True
        elif a is not None:
            def jacobian(x): #pylint: disable=E0102,E0213
                a.AssembleLinearization(x.vec)
                return a.mat
            self.jacobian = jacobian
            self.second_order = True
            self.jacobianMatType = jacobianMatType
    def setup(self, x0):
        '''
        This is method is used to setup the PETSc SNES object

        :arg x0: NGSolve grid function reppresenting the initial guess
        '''
        ngsGridFucntion = GridFunction(self.fes)
        pVec = self.vectorMapping.petscVec(ngsGridFucntion.vec)
        self.snes.setFunction(self.petscResidual, pVec)
        if self.objective is not False: self.snes.setObjective(self.petscObjective)
        if self.second_order:
            J, P, _ = self.snes.getJacobian()
            self.snes.setJacobian(self.petscJacobian, J, P)
        self.pvec0 = self.vectorMapping.petscVec(x0.vec)
    def solve(self, x0):
        '''
        This is method solves the non-linear problem

        :arg x0: NGSolve grid function reppresenting the initial guess
        '''
        self.setup(x0)
        self.snes.solve(None,self.pvec0)
        self.solutionGridFucntion = GridFunction(self.fes)
        self.vectorMapping.ngsVec(self.pvec0,ngsVec=self.solutionGridFucntion.vec)
        return self.solutionGridFucntion
    def petscResidual(self,snes,x,f):
        '''
        This is method is used to wrap the callback to the resiudal in
        a PETSc compatible way

        :arg snes: PETSc SNES object reppresenting the non-linear solver
        
        :arg x: current guess of the solution as a PETSc Vec

        :arg f: residual function as PETSc Vec
        '''
        assert isinstance(snes,PETSc.SNES)
        ngsGridFuction = GridFunction(self.fes)
        self.vectorMapping.ngsVec(x,ngsVec=ngsGridFuction.vec)
        ngsGridFuction = self.residual(ngsGridFuction)
        self.vectorMapping.petscVec(ngsGridFuction.vec, petscVec=f)

    def residual(x): #pylint: disable=E0102,E0213,E0202
        '''
        Callback to the residual of the non-linear problem
        
        :arg x: current guess of the solution as a PETSc Vec

        :return: the residual as an NGSolve grid function
        '''
        raise NotImplementedError("No residual has been implemented yet.")

    def petscObjective(self, snes,x):
        '''
        This is method is used to wrap the callback to the objetcive in
        a PETSc compatible way

        :arg snes: PETSc SNES object reppresenting the non-linear solver
        
        :arg x: current guess of the solution as a PETSc Vec

        :arg energy: energy as a PETSc Scalar
        '''
        assert isinstance(snes,PETSc.SNES)
        ngsGridFuction = GridFunction(self.fes)
        self.vectorMapping.ngsVec(x,ngsVec=ngsGridFuction.vec)
        return self.objective(ngsGridFuction)

    def objective(x): #pylint: disable=E0102,E0213,E0202
        '''
        Callback to the objective of the non-linear problem
        
        :arg x: current guess of the solution as a PETSc Vec

        :return: the energy
        '''
        raise NotImplementedError("No residual has been implemented yet.")

    def petscJacobian(self,snes,x,J,P):
        '''
        This is method is used to wrap the callback to the Jacobian in
        a PETSc compatible way

        :arg snes: PETSc SNES object reppresenting the non-linear solver
        
        :arg x: current guess of the solution as a PETSc Vec

        :arg J: Jacobian computed at x as a PETSc Mat

        :arg P: preconditioner for the Jacobian computed at x
                as a PETSc Mat
        '''
        assert isinstance(snes,PETSc.SNES)
        ngsGridFuction = GridFunction(self.fes)
        self.vectorMapping.ngsVec(x, ngsVec=ngsGridFuction.vec)
        mat = self.jacobian(ngsGridFuction)
        Matrix(mat,self.fes, petscMat=P, matType=self.jacobianMatType)
        Matrix(mat,self.fes, petscMat=J, matType=self.jacobianMatType)

    def jacobian(x): #pylint: disable=E0102,E0213,E0202
        '''
        Callback to the Jacobian of the non-linear problem
        
        :arg x: current guess of the solution as a PETSc Vec

        :return: the Jacobian as an NGSolve matrix
        '''
        raise NotImplementedError("No Jacobian has been implemented yet.")



"""
Convenience functions that are FiniteElement backend independent.
"""

import numpy as np
from petsc4py import PETSc
from scipy.spatial.distance import cdist
import numpy.typing as npt

__all__ = ["find_permutation"]

@PETSc.Log.EventDecorator()
def find_permutation(points_a:npt.NDArray[np.inexact], points_b:npt.NDArray[np.inexact],
                     tol:float=1e-5):
    """ Find all permutations between a list of two sets of points.

    Given two numpy arrays of shape (ncells, npoints, dim) containing
    floating point coordinates for each cell, determine each index
    permutation that takes `points_a` to `points_b`. Ie:
    ```
    permutation = find_permutation(points_a, points_b)
    assert np.allclose(points_a[permutation], points_b, rtol=0, atol=tol)
    ```
    """
    if points_a.shape != points_b.shape:
        raise ValueError("`points_a` and `points_b` must have the same shape.")

    p = [np.where(cdist(a, b).T < tol)[1] for a, b in zip(points_a, points_b)]
    try:
        permutation = np.array(p, ndmin=2)
    except ValueError as e:
        raise ValueError(
            "It was not possible to find a permutation for every cell"
            " within the provided tolerance"
        ) from e

    if permutation.shape != points_a.shape[0:2]:
        raise ValueError(
            "It was not possible to find a permutation for every cell"
            " within the provided tolerance"
        )

    return permutation



'''
This module contains all the functions related to the SLEPc eigenvalue
solver (EPS/PEP) interface for NGSolve
'''
from petsc4py import PETSc
try:
    from slepc4py import SLEPc
except ImportError:
    import warnings
    warnings.warn("Import Warning: it was not possible to import SLEPc")
    SLEPc = None

from ngsolve import GridFunction

from ngsPETSc import Matrix, VectorMapping
class EigenSolver():
    """
    This calss creates a SLEPc Eigen Problem Solver (EPS/PEP) from NGSolve
    variational problem pencil, i.e.
    a0(u,v)+lam*a1(u,v)+(lam^2)*a2(u,v)+ ... = 0
    Inspired by Firedrake Eigensolver class.

    :arg pencil: tuple containing the bilinear forms a: V x V -> K composing
    the pencil, e.g. (m,a) with a = BilinearForm(grad(u),grad(v)*dx) and
    m = BilinearForm(-1*u*v*dx)

    :arg fes: finite element space V

    :arg nev: number of requested eigenvalue

    :arg ncv: dimension of the internal subspace used by SLEPc,
    by Default by SLEPc.DECIDE

    :arg solverParameters: parameters to be passed to the KSP solver

    :arg optionsPrefix: special solver options prefix for this specific Krylov solver

    """
    if SLEPc is not None:
        def __init__(self, pencil, fes, nev, ncv=SLEPc.DECIDE, optionsPrefix=None,
                    solverParameters=None):
            self.comm = PETSc.COMM_WORLD.tompi4py()
            if not isinstance(pencil, tuple): pencil=tuple([pencil])
            self.penLength = len(pencil)
            self.fes = fes
            self.nev = nev
            self.ncv = ncv
            self.solverParameters = solverParameters
            self.optionsPrefix = optionsPrefix
            options_object = PETSc.Options()
            if solverParameters is not None:
                for optName, optValue in self.solverParameters.items():
                    options_object[optName] = optValue

            self.pencilMats = []
            self.pencilFlags = []
            for a in pencil:
                self.pencilFlags += [a.flags.ToDict()]
                self.pencilMats += [Matrix(a.Assemble().mat, fes).mat]
                self.pencilMats[-1].setOptionsPrefix(self.optionsPrefix)
                self.pencilMats[-1].setFromOptions()
            self.eps = None
            self.pep = None
            if self.penLength > 2:
                self.isEPS = False
            else:
                self.isEPS = True
                self.setUpEPS()

        def setUpEPS(self):
            '''
            This function setup a SLEPc EPS if the pencil has shape either (m,a)
            or is simply a single matrix. 
            '''
            self.eps = SLEPc.EPS().create()
            self.eps.setType(SLEPc.EPS.Type.KRYLOVSCHUR)
            if self.penLength == 1:
                flag0 = self.pencilFlags[0]
                if "symmetric" in flag0.keys():
                    if flag0["symmetric"]:
                        self.eps.setProblemType(SLEPc.EPS.ProblemType.HEP)
                    else:
                        self.eps.setProblemType(SLEPc.EPS.ProblemType.NHEP)
                else:
                    self.eps.setProblemType(SLEPc.EPS.ProblemType.NHEP)
                self.eps.setOperators(self.pencilMats[0])
            else:
                flag0 = self.pencilFlags[0]
                flag1 = self.pencilFlags[1]
                if "symmetric" in flag0.keys() and "symmetric" in flag1.keys():
                    if flag0["symmetric"] and flag1["symmetric"]:
                        self.eps.setProblemType(SLEPc.EPS.ProblemType.GHEP)
                    else:
                        self.eps.setProblemType(SLEPc.EPS.ProblemType.GNHEP)
                else:
                    self.eps.setProblemType(SLEPc.EPS.ProblemType.GNHEP)
                self.pencilMats[0].scale(-1)
                self.eps.setOperators(self.pencilMats[1], self.pencilMats[0])
            self.eps.setDimensions(self.nev, self.ncv)
            self.eps.setOptionsPrefix(self.optionsPrefix)
            self.eps.setFromOptions()

        def solve(self):
            '''
            This function solve the eigenprobelm
            '''
            self.eps.solve()
            self.nconv = self.eps.getConverged()
            if self.nconv == 0:
                raise RuntimeError("Did not converge any eigenvalues.")
            return self.nconv

        def view(self):
            '''
            This function setup display the information about SLEPc EPS/PEP
            '''
            self.eps.view()

        def eigenValue(self, i):
            '''
            This function return the eigenvalue of the eigenproblem

            :arg i: index of the eigenvalue we are intrested in.

            '''
            lam = None
            if self.isEPS:
                lam = self.eps.getEigenvalue(i)
            return lam

        def eigenFunction(self, i):
            '''
            This function return the eigenfunction of the eigenproblem

            :arg i: index of the eigenfunction we are intrested in.

            '''
            self.vecMap = VectorMapping(self.fes)
            eigenModeReal = GridFunction(self.fes)
            eigenModeImag = GridFunction(self.fes)
            eignModePETScReal = self.pencilMats[0].createVecLeft()
            eignModePETScImag = self.pencilMats[0].createVecLeft()
            if self.isEPS:
                self.eps.getEigenvector(i, eignModePETScReal, eignModePETScImag)
            self.vecMap.ngsVec(eignModePETScReal,eigenModeReal.vec)
            self.vecMap.ngsVec(eignModePETScImag,eigenModeImag.vec)
            return eigenModeReal, eigenModeImag

        def eigenValues(self, indeces):
            '''
            This function return the eigenvalues of the eigenproblem

            :arg indeces: indeces of the eigenvalues we are intrested in.

            '''
            lams = []
            if self.isEPS:
                for i in indeces:
                    lams = lams + [self.eps.getEigenvalue(i)]
            return lams

        def eigenFunctions(self, indeces):
            '''
            This function return a multidim with 
            the eigenfunctions of the eigenproblem

            :arg indeces: indeces of the eigenfunctions we are intrested in.

            '''
            self.vecMap = VectorMapping(self.fes)
            eigenModesReal = GridFunction(self.fes, multidim=self.nev)
            eigenModesImag = GridFunction(self.fes, multidim=self.nev)
            k = 0
            eignModePETScReal = self.pencilMats[0].createVecLeft()
            eignModePETScImag = self.pencilMats[0].createVecLeft()
            for i in indeces:
                if self.isEPS:
                    self.eps.getEigenvector(i, eignModePETScReal, eignModePETScImag)
                self.vecMap.ngsVec(eignModePETScReal,eigenModesReal.vecs[k])
                self.vecMap.ngsVec(eignModePETScImag,eigenModesImag.vecs[k])
                k += 1
            return eigenModesReal, eigenModesImag



'''
This module contains all the functions related to wrapping NGSolve meshes to
PETSc DMPlex using the petsc4py interface.
'''
import itertools
import numpy as np
from packaging.version import Version
from petsc4py import PETSc

import netgen.meshing as ngm
try:
    import ngsolve as ngs
except ImportError:
    class ngs:
        "dummy class"
        class comp:
            "dummy class"
            Mesh = type(None)

FACE_SETS_LABEL = "Face Sets"
CELL_SETS_LABEL = "Cell Sets"
EDGE_SETS_LABEL = "Edge Sets"

class MeshMapping:
    '''
    This class creates a mapping between Netgen/NGSolve meshes and PETSc DMPlex  

    :arg mesh: the mesh object, it can be either a Netgen/NGSolve mesh or a PETSc DMPlex

    :arg name: the name of to be assigned to the PETSc DMPlex, by default this is set to "Default"

    '''

    def __init__(self, mesh=None, comm=PETSc.COMM_WORLD.tompi4py(), name="Default"):
        self.name = name
        self.comm = comm
        if isinstance(mesh,(ngs.comp.Mesh,ngm.Mesh)):
            self.createPETScDMPlex(mesh)
        elif isinstance(mesh,PETSc.DMPlex):
            self.createNGSMesh(mesh)
        else:
            raise ValueError("Mesh format not recognised.")

    def createNGSMesh(self, plex):
        '''
        This function generate an NGSolve mesh from a PETSc DMPlex 

        :arg plex: the PETSc DMPlex to be converted in NGSolve mesh object

        '''
        self.petscPlex = plex
        ngMesh = ngm.Mesh(dim=plex.getCoordinateDim())
        self.ngMesh = ngMesh

        if plex.getDimension() == 2:
            coordinates = plex.getCoordinates().getArray().reshape([-1,2])
            self.ngMesh.AddPoints(coordinates)
            cStart,cEnd = plex.getHeightStratum(0)
            vStart, _ = plex.getHeightStratum(2)
            cells = []
            for i in range(cStart,cEnd):
                sIndex  = plex.getCone(i)
                if len(sIndex)==3:
                    S = list(itertools.chain.from_iterable([
                        list(plex.getCone(sIndex[k])-vStart) for k in range(len(sIndex))]))
                    cell = list(dict.fromkeys(S))
                    A = np.array([coordinates[cell[1]]-coordinates[cell[0]],
                                 coordinates[cell[2]]-coordinates[cell[1]]])
                    if np.linalg.det(A) > 0:
                        cells = cells + [cell]
                    else:
                        cells = cells + [[cell[0],cell[2],cell[1]]]
                else:
                    S = [plex.getCone(sIndex[k])-vStart for k in list(range(len(sIndex)))]
                    k = 0
                    cell = []
                    while k <  len(sIndex):
                        if S[k][0] == S[k-1][1]:
                            cell = cell+[S[k][0]]
                        else:
                            if k == 0  and S[k][0] not in list(S[k+1]):
                                S[-1] = np.array([S[-1][1],S[-1][0]])
                            else:
                                S[k] = np.array([S[k][1],S[k][0]])
                            cell = cell+[S[k][0]]
                        k = k+1
                    cells = cells + [cell]
            self.ngMesh.Add(ngm.FaceDescriptor(bc=1))
            self.ngMesh.AddElements(dim=2, index=1, data=np.asarray(cells, dtype=np.int32), base=0)
            for bcLabel in range(1,plex.getLabelSize(FACE_SETS_LABEL)+1):
                bcIndices = plex.getStratumIS("Face Sets",bcLabel).indices
                for j in bcIndices:
                    bcIndex = plex.getCone(j)-vStart
                    if len(bcIndex) == 2:
                        edge = ngm.Element1D([v+1 for v in bcIndex],index=bcLabel)
                        self.ngMesh.Add(edge)
        elif plex.getDimension() == 3:
            coordinates = plex.getCoordinates().getArray().reshape([-1,3])
            self.ngMesh.AddPoints(coordinates)
            cStart, cEnd = plex.getHeightStratum(0)
            vStart, _ = plex.getHeightStratum(3)
            cells = []
            for i in range(cStart,cEnd):
                fIndex  = plex.getCone(i)
                f1 = plex.getCone(fIndex[0])
                f2 = plex.getCone(fIndex[1])
                sIndex = [f1[0],f1[1],f1[2],f2[0],f2[1],f2[2]]
                if len(fIndex) == 4:
                    S = list(itertools.chain.from_iterable([
                        list(plex.getCone(sIndex[k])-vStart) for k in list(range(len(sIndex)))]))
                    cell = list(dict.fromkeys(S))
                    A = np.array([coordinates[cell[1]]-coordinates[cell[0]],
                                  coordinates[cell[2]]-coordinates[cell[1]],
                                  coordinates[cell[3]]-coordinates[cell[2]]])
                    if np.linalg.det(A) > 0:
                        cells = cells + [cell]
                    else:
                        cells = cells + [[cell[0],cell[1],cell[3],cell[2]]]
            #fd = self.ngmesh.Add(ngm.FaceDescriptor(bc=plex.getLabelSize(FACE_SETS_LABEL)+1))
            self.ngMesh.Add(ngm.FaceDescriptor(bc=plex.getLabelSize(FACE_SETS_LABEL)+1))
            self.ngMesh.AddElements(dim=3, index=plex.getLabelSize(FACE_SETS_LABEL)+1,
                                    data=np.asarray(cells,dtype=np.int32), base=0)
            for bcLabel in range(1,plex.getLabelSize(FACE_SETS_LABEL)+1):
                faces = []
                bcIndices = plex.getStratumIS("Face Sets",bcLabel).indices
                for j in bcIndices:
                    sIndex  = plex.getCone(j)
                    if len(sIndex)==3:
                        S = list(itertools.chain.from_iterable([
                            list(plex.getCone(sIndex[k])-vStart) for k in range(len(sIndex))]))
                        face = list(dict.fromkeys(S))
                        A = np.array([coordinates[face[1]]-coordinates[face[0]],
                                      coordinates[face[2]]-coordinates[face[1]],
                                      coordinates[face[0]]-coordinates[face[2]]])
                        eig = np.linalg.eig(A)[0].real
                        if eig[1]*eig[2] > 0:
                            faces = faces + [face]
                        else:
                            faces = faces + [[face[0],face[2],face[1]]]
                #fd = self.ngmesh.Add(ngm.FaceDescriptor(bc=bcLabel))
                self.ngMesh.Add(ngm.FaceDescriptor(bc=bcLabel))
                self.ngMesh.AddElements(dim=2, index=bcLabel,
                                        data=np.asarray(faces,dtype=np.int32), base=0)
        else:
            raise NotImplementedError("No implementation for dimension greater than 3.")

    def createPETScDMPlex(self, mesh):
        '''
        This function generate an PETSc DMPlex from a Netgen/NGSolve mesh object

        :arg plex: the Netgen/NGSolve mesh object to be converted into a PETSc
                   DMPlex.

        '''
        if isinstance(mesh,ngs.comp.Mesh):
            self.ngMesh = mesh.ngmesh
        else:
            self.ngMesh = mesh
        comm = self.comm
        if self.ngMesh.dim == 3:
            if comm.rank == 0:
                V = self.ngMesh.Coordinates()
                T = self.ngMesh.Elements3D().NumPy()["nodes"]

                surfMesh, dim = False, 3
                if len(T) == 0:
                    surfMesh, dim = True, 2
                    T = self.ngMesh.Elements2D().NumPy()["nodes"]

                if Version(np.__version__) >= Version("2.2"):
                    T = np.trim_zeros(T, "b", axis=1).astype(np.int32) - 1
                else:
                    T = (
                        np.array(
                            [list(np.trim_zeros(a, "b")) for a in list(T)],
                            dtype=np.int32,
                        )
                        - 1
                    )

                plex = PETSc.DMPlex().createFromCellList(dim, T, V, comm=comm)
                plex.setName(self.name)
                vStart, _ = plex.getDepthStratum(0)
                if surfMesh:
                    for e in self.ngMesh.Elements1D():
                        join = plex.getJoin([vStart+v.nr-1 for v in e.vertices])
                        plex.setLabelValue(FACE_SETS_LABEL, join[0], int(e.surfaces[1]))
                else:
                    for e in self.ngMesh.Elements2D():
                        join = plex.getFullJoin([vStart+v.nr-1 for v in e.vertices])
                        plex.setLabelValue(FACE_SETS_LABEL, join[0], int(e.index))
                    for e in self.ngMesh.Elements1D():
                        join = plex.getJoin([vStart+v.nr-1 for v in e.vertices])
                        plex.setLabelValue(EDGE_SETS_LABEL, join[0], int(e.index))
                self.petscPlex = plex
            else:
                plex = PETSc.DMPlex().createFromCellList(3,
                                                        np.zeros((0, 4), dtype=np.int32),
                                                        np.zeros((0, 3), dtype=np.double),
                                                        comm=comm)
                self.petscPlex = plex
        elif self.ngMesh.dim == 2:
            if comm.rank == 0:
                V = self.ngMesh.Coordinates()
                T = self.ngMesh.Elements2D().NumPy()["nodes"]
                T = np.array([list(np.trim_zeros(a, 'b')) for a in list(T)])-1
                plex = PETSc.DMPlex().createFromCellList(2, T, V, comm=comm)
                plex.setName(self.name)
                vStart, _ = plex.getDepthStratum(0)   # vertices
                for e in self.ngMesh.Elements1D():
                    join = plex.getJoin([vStart+v.nr-1 for v in e.vertices])
                    plex.setLabelValue(FACE_SETS_LABEL, join[0], int(e.index))
                if not (1 == self.ngMesh.Elements2D().NumPy()["index"]).all():
                    for e in self.ngMesh.Elements2D():
                        join = plex.getFullJoin([vStart+v.nr-1 for v in e.vertices])
                        plex.setLabelValue(CELL_SETS_LABEL, join[0], int(e.index))

                self.petscPlex = plex
            else:
                plex = PETSc.DMPlex().createFromCellList(2,
                                                        np.zeros((0, 3), dtype=np.int32),
                                                        np.zeros((0, 2), dtype=np.double),
                                                        comm=comm)
                self.petscPlex = plex



'''
This module contains all the function and class needed to wrap a PETSc Preconditioner in NGSolve
'''
from petsc4py import PETSc

from ngsolve import BaseMatrix, comp
from ngsPETSc import Matrix, VectorMapping

class PETScPreconditioner(BaseMatrix):
    '''
    This class creates a Netgen/NGSolve BaseMatrix corresponding to a PETSc PC  

    :arg mat: NGSolve Matrix one would like to build the PETSc preconditioner for.

    :arg freeDofs: not constrained degrees of freedom of the finite element space over
    which the BilinearForm corresponding to the matrix is defined.

    :arg solverParameters: parameters to be passed to the KSP solver

    :arg optionsPrefix: special solver options prefix for this specific Krylov solver

    :arg matType: type of sparse matrix, i.e. PETSc sparse: aij, 
    MKL sparse: mklaij or CUDA: aijcusparse

    '''
    def __init__(self, mat, freeDofs, solverParameters=None, optionsPrefix="", nullspace=None,
                 matType="aij"):
        BaseMatrix.__init__(self)
        if hasattr(solverParameters, "ToDict"):
            solverParameters = solverParameters.ToDict()
        self.ngsMat = mat
        if hasattr(self.ngsMat, "row_pardofs"):
            dofs = self.ngsMat.row_pardofs
        else:
            dofs = None
        self.vecMap = VectorMapping((dofs,freeDofs,{"bsize":self.ngsMat.local_mat.entrysizes}))
        petscMat = Matrix(self.ngsMat, (dofs, freeDofs, None), matType).mat
        self.lgmap = petscMat.getLGMap()[0]
        if nullspace is not None:
            if nullspace.near:
                petscMat.mat.setNearNullSpace(nullspace.nullspace)
        self.petscPreconditioner = PETSc.PC().create(comm=petscMat.getComm())
        self.petscPreconditioner.setOperators(petscMat)
        options_object = PETSc.Options()
        if solverParameters is not None:
            for optName, optValue in solverParameters.items():
                options_object[optName] = optValue
        self.petscPreconditioner.setOptionsPrefix(optionsPrefix)
        self.petscPreconditioner.setFromOptions()
        self.petscPreconditioner.setUp()
        self.petscVecX, self.petscVecY = petscMat.createVecs()
    def Shape(self):
        '''
        Shape of the BaseMatrix

        '''
        return self.ngsMat.shape

    def CreateVector(self,col):
        '''
        Create vector corresponding to the matrix

        :arg col: True if one want a column vector

        '''
        return self.ngsMat.CreateVector(not col)

    def Mult(self,x,y):
        '''
        BaseMatrix multiplication Ax = y
        :arg x: vector we are multiplying
        :arg y: vector we are storeing the result in

        '''
        self.vecMap.petscVec(x,self.petscVecX)
        self.petscPreconditioner.apply(self.petscVecX, self.petscVecY)
        self.vecMap.ngsVec(self.petscVecY, y)

    def MultTrans(self,x,y):
        '''
        BaseMatrix multiplication A^T x = y
        :arg x: vector we are multiplying
        :arg y: vector we are storeing the result in

        '''
        self.vecMap.petscVec(x,self.petscVecX)
        self.petscPreconditioner.applyTranspose(self.petscVecX, self.petscVecY)
        self.vecMap.ngsVec(self.petscVecY, y)

    def setActingDofs(self, dofs):
        '''
        Set the acting dofs of the preconditioner
        :arg dofs: dofs that the preconditioner is acting on
        '''
        self.actingDofs = dofs

def createPETScPreconditioner(mat, freeDofs, solverParameters):
    '''
    Create PETSc PC that can be accessed by NGSolve.
    :arg mat: NGSolve Matrix one would like to build the PETSc preconditioner for.

    :arg freeDofs: not constrained degrees of freedom of the finite element space over
    which the BilinearForm corresponding to the matrix is defined.

    :arg solverParameters: parameters to be passed to the KSP solver

    '''
    return PETScPreconditioner(mat, freeDofs, solverParameters)


class ASMPreconditioner(PETScPreconditioner):
    '''
    This class creates a Netgen/NGSolve BaseMatrix corresponding to a PETSc ASM PC
    '''
    def __init__(self, mat, freeDofs, solverParameters=None, optionsPrefix="",
                 nullspace=None, matType="aij", blocks=None):

        if "sub_pc_type" not in solverParameters:
            solverParameters["sub_pc_type"] = "lu"
        if "sub_pc_factor_mat_ordering_type"  not in solverParameters:
            solverParameters["sub_pc_factor_mat_ordering_type"] = "natural"
        super().__init__(mat, freeDofs, solverParameters, optionsPrefix, nullspace, matType)
        self.asmpc = None
        if blocks is not None:
            if len (blocks) == 0:
                self.ises = [PETSc.IS().createGeneral([0], comm=PETSc.COMM_SELF)]
            else:
                if self.petscPreconditioner.getComm().size > 1:
                    self.ises = [self.lgmap.applyIS(PETSc.IS().createGeneral(list(block),
                                 comm=PETSc.COMM_SELF)) for block in blocks]
                else:
                    self.ises = [PETSc.IS().createGeneral(list(block),
                                 comm=PETSc.COMM_SELF) for block in blocks]
            self.asmpc = PETSc.PC().create(comm=self.petscPreconditioner.getComm())
            self.asmpc.incrementTabLevel(1, parent=self.petscPreconditioner)
            self.asmpc.setOptionsPrefix(optionsPrefix)
            self.asmpc.setFromOptions()
            self.asmpc.setOperators(*self.petscPreconditioner.getOperators())
            self.asmpc.setType(self.asmpc.Type.ASM)
            self.asmpc.setASMType(PETSc.PC.ASMType.BASIC)
            self.asmpc.setASMLocalSubdomains(len(self.ises), self.ises)
            self.asmpc.setUp()

    def Mult(self,x,y):
        '''
        BaseMatrix multiplication Ax = y
        :arg x: vector we are multiplying
        :arg y: vector we are storeing the result in

        '''
        self.vecMap.petscVec(x,self.petscVecX)
        if self.asmpc is not None:
            self.asmpc.apply(self.petscVecX, self.petscVecY)
        else:
            self.petscPreconditioner.apply(self.petscVecX, self.petscVecY)
        self.vecMap.ngsVec(self.petscVecY, y)

    def MultTrans(self,x,y):
        '''
        BaseMatrix multiplication A^T x = y
        :arg x: vector we are multiplying
        :arg y: vector we are storeing the result in

        '''
        self.vecMap.petscVec(x,self.petscVecX)
        if self.asmpc is not None:
            self.asmpc.applyTranspose(self.petscVecX, self.petscVecY)
        else:
            self.petscPreconditioner.applyTranspose(self.petscVecX, self.petscVecY)
        self.vecMap.ngsVec(self.petscVecY, y)

comp.RegisterPreconditioner ("PETScPC", createPETScPreconditioner)



'''
ngsPETSc is a NGSolve/Netgen interface to PETSc
'''
from ngsPETSc.plex import *

__all__ = []

#Firedrake
try:
    import firedrake
except ImportError:
    firedrake = None

if firedrake:
    from ngsPETSc.utils.firedrake.meshes import *
    from ngsPETSc.utils.firedrake.hierarchies import *
    __all__ = __all__ + ["FiredrakeMesh", "NetgenHierarchy"]

#FEniCSx
try:
    import dolfinx
except ImportError:
    dolfinx = None

if dolfinx:
    from ngsPETSc.utils.fenicsx import *

#Netgen
try:
    import ngsolve
except ImportError:
    ngsolve = None

if ngsolve:
    from ngsPETSc.mat import *
    from ngsPETSc.vec import *
    from ngsPETSc.nullspace import *
    from ngsPETSc.pc import *
    from ngsPETSc.ksp import *
    from ngsPETSc.snes import *
    from ngsPETSc.eps import *
    __all__ = __all__ + ["Matrix","VectorMapping","MeshMapping",
                         "KrylovSolver","EigenSolver","NullSpace",
                         "PETScPreconditioner", "NonLinearSolver"]

VERSION = "0.0.5"



'''
This module contains all the functions related to the NGSolve vector -
PETSc vector mapping using the petsc4py interface.
'''
import numpy as np

from petsc4py import PETSc

from ngsolve import la, FESpace


class VectorMapping:
    '''
    This class creates a mapping between a PETSc vector and NGSolve
    vectors

    :arg parDescr: the finite element space for the vector or tuple (dofs, freeDofs)
    :arg prefix: prefix for PETSc options

    '''
    def __init__(self, parDescr, prefix='ngs_'):
        if isinstance(parDescr, FESpace):
            dofs = parDescr.ParallelDofs()
            freeDofs = parDescr.FreeDofs()
            comm = dofs.comm.mpi4py
        else:
            dofs, freeDofs, dofsInfo = parDescr
            if dofs is not None:
                comm = dofs.comm.mpi4py
            else:
                ### create suitable dofs
                comm = PETSc.COMM_SELF
                dofs = type('', (object,), {'entrysize':dofsInfo["bsize"][0]})()

        self.dofs = dofs
        bsize = dofs.entrysize
        locfree = np.flatnonzero(freeDofs).astype(PETSc.IntType)
        isetlocfree = PETSc.IS().createBlock(indices=locfree,
                                             bsize=bsize, comm=PETSc.COMM_SELF)
        nloc = len(freeDofs)

        nglob = len(locfree)
        if comm.Get_size() > 1:
            globnums, nglob = dofs.EnumerateGlobally(freeDofs)
            globnums = np.array(globnums, dtype=PETSc.IntType)[freeDofs]
            iset = PETSc.IS().createBlock(indices=globnums, bsize=bsize, comm=comm)
        else:
            iset = PETSc.IS().createBlock(indices=np.arange(nglob,dtype=PETSc.IntType),
                                          bsize=bsize, comm=comm)

        self.sVec = PETSc.Vec().create(comm=PETSc.COMM_SELF)
        self.sVec.setSizes(nloc*bsize,bsize=bsize)
        self.sVec.setOptionsPrefix(prefix)
        self.sVec.setFromOptions()

        self.pVec = PETSc.Vec().create(comm=comm)
        self.pVec.setSizes(nglob*bsize,bsize=bsize)
        self.pVec.setBlockSize(bsize)
        self.pVec.setOptionsPrefix(prefix)
        self.pVec.setFromOptions()

        self.ngsToPETScScat = PETSc.Scatter().create(self.sVec, isetlocfree,
                                                     self.pVec, iset)

    def petscVec(self, ngsVec, petscVec=None):
        '''
        This function generate a PETSc vector from a NGSolve vector

        :arg ngsVec: the NGSolve vector
        :arg petscVec: the PETSc vector to be loaded with NGSolve
        vector, if None new PETSc vector is generated, by deafault None.

        '''
        if hasattr(ngsVec, "nblocks"):
            if petscVec is None:
                petscVec = self.pVec.duplicate()
            indexSum = 0
            for i in range(ngsVec.nblocks):
                ngsVec[i].Distribute()
                I = list(range(indexSum, indexSum + ngsVec[i].size))
                petscVec.setValues(I, ngsVec[i].FV().NumPy())
                indexSum += ngsVec[i].size
        else:
            if petscVec is None:
                petscVec = self.pVec.duplicate()
            ngsVec.Distribute()
            self.sVec.placeArray(ngsVec.FV().NumPy())
            petscVec.set(0)
            self.ngsToPETScScat.scatter(self.sVec, petscVec, addv=PETSc.InsertMode.ADD,
                                        mode=PETSc.ScatterMode.FORWARD)
            self.sVec.resetArray()
        return petscVec

    def ngsVec(self, petscVec, ngsVec=None):
        '''
        This function generate a NGSolve vector from a PETSc vector

        :arg petscVec: the PETSc vector
        :arg ngsVec: the NGSolve vector vector to be loaded with PETSc
        vector, if None new PETSc vector is generated, by deafault None.

        '''
        if ngsVec is None:
            ngsVec = la.CreateParallelVector(self.dofs,la.PARALLEL_STATUS.CUMULATED)
        ngsVec[:] = 0
        if hasattr(ngsVec, "nblocks"):
            indexSum = 0
            for i in range(ngsVec.nblocks):
                I = list(range(indexSum, indexSum + ngsVec[i].size))
                ngsVec[i].FV().NumPy()[:] = petscVec.getValues(I)
                indexSum += ngsVec[i].size
        else:
            self.sVec.placeArray(ngsVec.FV().NumPy())
            self.ngsToPETScScat.scatter(petscVec, self.sVec, addv=PETSc.InsertMode.INSERT,
                                        mode=PETSc.ScatterMode.REVERSE)
            self.sVec.resetArray()
        return ngsVec
